# coding: utf-8
"""
Molecular line strength calculation

References:
    Istvan Kovacs, Rotational Structure in the spectra of diatomic molecules.
    American Elsevier, 1969

    https://en.wikipedia.org/wiki/Multiplicity_(chemistry)

    Bruno Castilho's work on ATMOS machine
"""


import math
import pyfant


__all__ = ["kovacs_toolbox", "NoLineStrength", "NO_LINE_STRENGTH"]


def kovacs_toolbox(molconsts, flag_normalize=None):
    """Factory function that returns a MultiplicityToolbox descendant appropriate to molconsts"""

    C = [_LSTSinglet0, _LSTSinglet1, _LSTDoublet0, _LSTDoublet1, _LSTTriplet0, _LSTTriplet1]

    absDeltaLambda = abs(molconsts["from_spdf"]-molconsts["to_spdf"])

    for cls in C:
        if (cls.absDeltaLambda == "all" or cls.absDeltaLambda == absDeltaLambda) and \
           cls.multiplicityl == molconsts["from_mult"] and \
           cls.multiplicity2l == molconsts["to_mult"]:
            return cls(molconsts, flag_normalize)

    raise ValueError("Could not find a suitable class for given molecular constants "
     "(I need abs(\u0394\u039B)={}; multiplicity'={}; multiplicity''={})".format(
     int(absDeltaLambda), int(molconsts["from_mult"]), int(molconsts["to_mult"])))


class NoLineStrength(Exception):
    pass


# Code to signal that there is no Hönl-London factor for combination (vl, v2l, J, branch)
# (will be appear instead of HLF)
NO_LINE_STRENGTH = -999999


def _quanta_to_branch_singlet(Jl, J2l, spinl=None, spin2l=None):
    """
    Singlet only has branches P/Q/R

    Args:
        Jl: J upper or J'
        J2l: J lower or  J''
        spinl: ignored.
        spin2l: ignored

    Note: spinl, spin2l are present only to keep the interface consistent with
          .(singlet/doublet).quanta_to_branch()

    Returns: branch letter: "P"/"Q"/"R"
    """
    if Jl > J2l:
        return "R"
    elif Jl == J2l:
        return "Q"
    return "P"


def _quanta_to_branch_same_multiplicity(Jl, J2l, spinl=None, spin2l=None):
    """
    Converts quanta to branch for cases where the initial and final states have the same multiplicity, except singlets

    Args:
        Jl: J upper or J'
        J2l: J lower or  J''
        spinl:
        spin2l: (cannot be None)

    Returns:
        str: if spinl is None or equals spin2l, (letter)(number),
             otherwise (letter)(number)(number)

    >>> _quanta_to_branch_same_multiplicity(10.5, 11.5, spin2l=1)
    'P1'
    >>> _quanta_to_branch_same_multiplicity(11.5, 10.5, spin2l=2)
    'R2'
    >>> _quanta_to_branch_same_multiplicity(10.5, 10.5, spin2l=1)
    'Q1'
    """

    if Jl > J2l:
        br = "R"
    elif Jl == J2l:
        br =  "Q"
    else:
        br = "P"

    if spinl is None:
        ret = "{}{:1d}".format(br, spin2l)
    else:
        if spin2l is None:
          raise ValueError("spin2l cannot be None")

        if spinl == spin2l:
            ret = "{}{:1d}".format(br, spin2l)
        else:
            ret = "{}{:1d}{:1d}".format(br, spinl, spin2l)
    return ret


class _LineStrengthToolbox(object):
    """
    Tools: Kovacz' "Line strength" and and quanta-to-branch conversion

    This class has descendants to handle specific cases. Basically, each descendant implements one
    table in Kovacs Chapter 3

    Usage:

        - Line strength: method get_sj()
        - convert quantum information to branch: method quanta_to_branch()
    """

    # # These values must be set at descendants
    # This allows for
    #     - automatically finding a suitable MultiplicityToolbox descendant given a system (e.g. A 2 Sigma - X 2 Pi)
    # and - error checking
    #
    # Possible values: "all", (Delta Lambda = LAML - LAM2L)
    #
    absDeltaLambda = None
    # 1 for singlet; 2 for doublet; 3 for triplet
    multiplicityl = None
    multiplicity2l = None
    # Method to convert quantum information to branch. **Use staticmethod()**!
    quanta_to_branch = None

    @property
    def dict_sj(self):
        return self._dict_sj

    def __init__(self, molconsts, flag_normalize=True):
        if not isinstance(molconsts, pyfant.MolConsts):
            raise TypeError("molconsts must be a MolConsts")
        self._molconsts = molconsts

        LAML = molconsts["from_spdf"]
        LAM2L = molconsts["to_spdf"]
        S = molconsts.get_S2l()

        if self.absDeltaLambda != "all" and abs(LAML - LAM2L) != self.absDeltaLambda:
                raise ValueError("Invalid Delta Lambda {}. abs(Delta Lambda) must be {}". \
                             format(LAML-LAM2L, repr(self.absDeltaLambda)))

        if 2*S+1 != self.multiplicity2l:
            raise ValueError("**Sanity check fail**: Class {} expects multiplicity2l = 2*S+1 = {}, got {} instead". \
                             format(self.__class__.__name__, self.multiplicity2l, 2*S+1))

        self._dict_sj = {}

        #
        self._flag_normalize = flag_normalize

    def populate(self, vl, v2l, J):
        """Populates self.dict_sj with keys (vl, v2l, J, (all branches))"""
        data = self._get_populate_data(vl, v2l, J)
        self.__update_with_data(vl, v2l, J, data)

    def get_sj(self, vl, v2l, J, branch):
        """
        Returns Kovacz' "Line Strength", normalized or not

        Args:
            vl:
            v2l:
            J: **integer**
            branch: example: "P1"
        """

        J = int(J)

        key = (vl, v2l, J, branch)

        if key not in self._dict_sj:
            data = self._get_populate_data(vl, v2l, J)
            self.__update_with_data(vl, v2l, J, data)

        value = self._dict_sj[key]

        if value == NO_LINE_STRENGTH:
            raise NoLineStrength(
             "Cannot calculate line strength for (vl={}, v2l={}, J={}, branch='{}')".format(*key))

        return value

    def __update_with_data(self, vl, v2l, J, data):
        normalization_factor = 1.
        if self._flag_normalize:
            cc = self._molconsts
            # TODO 20230415 Was 2, I put 1
            normalization_factor = 1. / ((2 * cc.get_S2l() + 1) * (2 * J + 1) * (2 - cc.get_deltak()))
            # normalization_factor = 2. / ((2 * cc.get_S2l() + 1) * (2 * J + 1) * (2 - cc.get_deltak()))

        for branch, function in data:
            try:
                value = function(J) * normalization_factor

                if isinstance(value, complex):
                    value = NO_LINE_STRENGTH

            except ValueError:
                # Here we expect "ValueError: math domain error" (sqrt of negative number)
                value = NO_LINE_STRENGTH
            except ZeroDivisionError:
                value = NO_LINE_STRENGTH
            self._dict_sj[(vl, v2l, J, branch)]  = value


    def _get_populate_data(self, vl, v2l, J):
        """Must be inherited and populate self with all branches for given (vl, v2l, J)"""
        raise NotImplementedError()


class _LSTSinglet(_LineStrengthToolbox):
    absDeltaLambda = "all"
    multiplicityl = 1
    multiplicity2l = 1
    quanta_to_branch = staticmethod(_quanta_to_branch_singlet)


class _LSTSinglet0(_LSTSinglet):
    absDeltaLambda = 0
    multiplicityl = 1
    multiplicity2l = 1
    quanta_to_branch = staticmethod(_quanta_to_branch_singlet)

    def _get_populate_data(self, vl, v2l, J):
        cc = self._molconsts
        LAML = cc["from_spdf"]
        LAM2L = cc["to_spdf"]


        # Kovacs Table 3.2

        return (
         ("P", lambda J: (J+LAML)*(J-LAML)/J),
         ("Q", lambda J: LAML**2*(2*J+1)/J/(J+1)),
         ("R", lambda J: (J+LAML+1)*(J-LAML+1)/(J+1)),
        )


class _LSTSinglet1(_LSTSinglet):
    absDeltaLambda = 1
    multiplicityl = 1
    multiplicity2l = 1
    quanta_to_branch = staticmethod(_quanta_to_branch_singlet)

    def _get_populate_data(self, vl, v2l, J):
        cc = self._molconsts
        LAML = cc["from_spdf"]
        LAM2L = cc["to_spdf"]

        LMIN = min(LAML, LAM2L)

        # Kovacz table 3.4
        _P = lambda J: (J - LMIN - 1) * (J - LMIN) / (2 * J)
        _Q = lambda J: (J - LMIN) * (J + LMIN + 1) * (2 * J + 1) / (2 * J * (J + 1))
        _R = lambda J: (J + LMIN + 1) * (J + LMIN + 2)  / (2 * (J + 1))

        # Resolves the Delta Lambda = LAML - LAM2L

        if LAML > LAM2L:
            return (
             ("P", _P),
             ("Q", _Q),
             ("R", _R),
            )

        elif LAML < LAM2L:
            return (
             ("P", lambda J: _R(J-1)),
             ("Q", _Q),
             ("R", lambda J: _P(J+1)),
            )


class _LSTDoublet(_LineStrengthToolbox):
    """
    Line strengths for doublets (any Delta Lambda)

    Formulas: Formula 2.1.3-6 (p61)
    """

    multiplicityl = 2
    multiplicity2l = 2
    quanta_to_branch = staticmethod(_quanta_to_branch_same_multiplicity)

    def _get_strengths(self, J, LAML, LAM2L, UPLUSL, UMINUSL,CPLUSL, CMINUSL, UPLUS2L,
                       UMINUS2L, CPLUS2L, CMINUS2L, YL, Y2L,):
        raise NotImplementedError()

    def _get_populate_data(self, vl, v2l, J):
        cc = self._molconsts
        LAML = cc["from_spdf"]
        LAM2L = cc["to_spdf"]
        AL = cc["statel_A"]
        AEL = cc["statel_alpha_e"]
        BEL = cc["statel_B_e"]
        A2L = cc["state2l_A"]
        AE2L = cc["state2l_alpha_e"]
        BE2L = cc["state2l_B_e"]


        BL = BEL - AEL * (vl + 0.5)
        B2L = BE2L - AE2L * (v2l + 0.5)

        YL = AL / BL
        Y2L = A2L / B2L

        # p61: “[...] Y= A/B. The term is called normal if A > 0, and inverted if A < 0”
        # p126: “in the case of inverted terms, Y is to be substituted with negative sign”
        if YL < 0:
            YL = -YL
        if Y2L < 0:
            Y2L = -Y2L

        # Kovacs formula 2.1.3-6
        UPLUSL = lambda J: ((LAML**2.)*YL*(YL-4) + 4*((J+0.5)**2.))**0.5 + LAML*(YL-2)
        UMINUSL= lambda J: ((LAML**2.)*YL*(YL-4) + 4*((J+0.5)**2.))**0.5 - LAML*(YL-2)
        CPLUSL = lambda J: 0.5*((UPLUSL(J)**2.) + 4*(((J+0.5)**2.) - LAML**2.))
        CMINUSL = lambda J: 0.5*((UMINUSL(J)**2.) + 4*(((J+0.5)**2.) - LAML**2.))
        UPLUS2L = lambda J: ((LAM2L**2.)*Y2L*(Y2L-4) + 4*((J+0.5)**2.))**0.5 + LAM2L*(Y2L-2)
        UMINUS2L = lambda J: ((LAM2L**2.)*Y2L*(Y2L-4) + 4*((J+0.5)**2.))**0.5 - LAM2L*(Y2L-2)
        CPLUS2L = lambda J: 0.5*((UPLUS2L(J)**2.) + 4*(((J+0.5)**2.)-LAM2L**2.))
        CMINUS2L = lambda J: 0.5*((UMINUS2L(J)**2.)+4*(((J+0.5)**2.)-LAM2L**2.))

        return self._get_strengths(J, LAML, LAM2L, UPLUSL, UMINUSL, CPLUSL, CMINUSL, UPLUS2L,
                                   UMINUS2L, CPLUS2L, CMINUS2L, YL, Y2L,)


class _LSTDoublet0(_LSTDoublet):
    """
    Line strengths for (doublet, Delta Lambda = 0)

    Formulas: Table 3.6 (p127)
    """

    absDeltaLambda = 0

    def _get_strengths(self, J, LAML, LAM2L, UPLUSL, UMINUSL,CPLUSL, CMINUSL, UPLUS2L,
                       UMINUS2L, CPLUS2L, CMINUS2L, YL, Y2L,):

        LAM = LAML  # remember, both LAML, LAM2L are the same

        _P1 = lambda J: \
            (((J-LAM-0.5)*(J+LAM+0.5))/(4*J*CMINUSL(J-1)*CMINUS2L(J)))* \
            ((UMINUSL(J-1)*UMINUS2L(J) + 4*(J-LAM+0.5)*(J+LAM-0.5))**2.)

        _Q1 = lambda J: \
            ((J+0.5)/(2*J*(J+1)*CMINUSL(J)*CMINUS2L(J)))* \
            (((LAM+0.5)*UMINUSL(J)*UMINUS2L(J) + 4*(LAM-0.5)*(J-LAM+0.5)*(J+LAM+0.5))**2.)

        _R1 = lambda J: \
            (((J-LAM+0.5)*(J+LAM+1.5))/(4*(J+1)*CMINUSL(J+1)*CMINUS2L(J)))* \
            ((UMINUSL(J+1)*UMINUS2L(J) + 4*(J-LAM+1.5)*(J+LAM+0.5))**2.)

        _P21 = lambda J: \
            (((J-LAM-0.5)*(J+LAM+0.5))/(4*J*CPLUSL(J-1)*CMINUS2L(J)))* \
            ((UPLUSL(J-1)*UMINUS2L(J) - 4*(J-LAM+0.5)*(J+LAM-0.5))**2)

        _Q21 = lambda J: \
            (((J+0.5))/(2*J*(J+1)*CPLUSL(J)*CMINUS2L(J)))* \
            ((LAM+0.5)*(UPLUSL(J)*UMINUS2L(J) - 4*(LAM-0.5)*(J-LAM+0.5)*(J+LAM+0.5))**2.)

        _R21 = lambda J: \
            (((J-LAM+0.5)*(J+LAM+1.5))/(4*(J+1)*CPLUSL(J+1)*CMINUS2L(J)))* \
            ((UPLUSL(J+1)*UMINUS2L(J) - 4*(J-LAM+1.5)*(J+LAM+0.5))**2.)

        _P12 = lambda J: \
            (((J-LAM-0.5)*(J+LAM+0.5))/(4*J*CMINUSL(J-1)*CPLUS2L(J)))* \
            ((UMINUSL(J-1)*UPLUS2L(J) - 4*(J-LAM+0.5)*(J+LAM-0.5))**2.)

        _Q12 = lambda J: \
            (((J+0.5))/(2*J*(J+1)*CMINUSL(J)*CPLUS2L(J)))*\
            ((LAM+0.5)*(UMINUSL(J)*UPLUS2L(J) - 4*(LAM-0.5)*(J-LAM+0.5)*(J+LAM+0.5))**2.)

        _R12 = lambda J: \
            (((J-LAM+0.5)*(J+LAM+1.5))/(4*(J+1)*CMINUSL(J+1)*CPLUS2L(J)))* \
            ((UMINUSL(J+1)*UPLUS2L(J) - 4*(J-LAM+1.5)*(J+LAM+0.5))**2.)

        _P2 = lambda J: \
            (((J-LAM-0.5)*(J+LAM+0.5))/(4*J*CPLUSL(J-1)*CPLUS2L(J)))* \
            ((UPLUSL(J-1)*UPLUS2L(J) + 4*(J-LAM+0.5)*(J+LAM-0.5))**2.)

        _Q2 = lambda J: \
            (((J+0.5))/(2*J*(J+1)* CPLUSL(J)*CPLUS2L(J)))* \
            ((LAM+0.5)*(UPLUSL(J)*UPLUS2L(J) + 4*(LAM-0.5)*(J-LAM+0.5)*(J+LAM+0.5))**2.)

        _R2 = lambda J: \
            (((J-LAM+0.5)*(J+LAM+1.5))/(4*(J+1)*CPLUSL(J+1)*CPLUS2L(J)))* \
            ((UPLUSL(J+1)*UPLUS2L(J) + 4*(J-LAM+1.5)*(J+LAM+0.5))**2.)

        return (
         ("P1", _P1),
         ("Q1", _Q1),
         ("R1", _R1),
         ("P21", _P21),
         ("Q21", _Q21),
         ("R21", _R21),
         ("P12", _P12),
         ("Q12", _Q12),
         ("R12", _R12),
         ("P2", _P2),
         ("Q2", _Q2),
         ("R2", _R2),
        )


class _LSTDoublet1(_LSTDoublet):
    """
    Line strengths for (doublet, Delta Lambda = +-1)

    Formulas: Table 3.7 (p130)
    """

    absDeltaLambda = 1

    def _get_strengths(self, J, LAML, LAM2L, UPLUSL, UMINUSL,CPLUSL, CMINUSL, UPLUS2L,
                       UMINUS2L, CPLUS2L, CMINUS2L, YL, Y2L,):

        # Original comment:
        #
        #     usa-se o menor valor de Lambda nas formulas abaixo
        #
        # p129: "[...] where the explicit Lambda values always mean the smaller of the two Lambda's
        #        involved in the transition"
        LMIN = min(LAML, LAM2L)

        _P1 = lambda J: \
            (((J-LMIN-1.5)*(J-LMIN-0.5))/(8*J*CMINUSL(J-1)*
            CMINUS2L(J)))*((UMINUSL(J-1)*UMINUS2L(J) +
            4*(J-LMIN+0.5)*(J+LMIN+0.5))**2.)

        _Q1 = lambda J: \
            (((J-LMIN-0.5)*(J+0.5)*(J+LMIN+1.5))/(4*J*(J+1)*
            CMINUSL(J)*CMINUS2L(J)))*((UMINUSL(J)*UMINUS2L(J) +
            4*(J-LMIN+0.5)*(J+LMIN+0.5))**2.)


        _R1 = lambda J: \
            (((J+LMIN+1.5)*(J+LMIN+2.5))/(8*(J+1)*CMINUSL(J+1)*
            CMINUS2L(J)))*((UMINUSL(J+1)*UMINUS2L(J) +
            4*(J-LMIN+0.5)*(J+LMIN+0.5))**2.)

        _P21 = lambda J: \
            (((J-LMIN-1.5)*(J-LMIN-0.5))/(8*J*CPLUSL(J-1)*
            CMINUS2L(J)))*((UPLUSL(J-1)*UMINUS2L(J) -
            4*(J-LMIN+0.5)*(J+LMIN+0.5))**2.)

        _Q21 = lambda J: \
            (((J-LMIN-0.5)*(J+0.5)*(J+LMIN+1.5))/(4*J*(J+1)*
            CPLUSL(J)*CMINUS2L(J)))*((UPLUSL(J)*UMINUS2L(J) -
            4*(J-LMIN+0.5)*(J+LMIN+0.5))**2.)


        _R21 = lambda J: \
            (((J+LMIN+1.5)*(J+LMIN+2.5))/(8*(J+1)*CPLUSL(J+1)*
            CMINUS2L(J)))*((UPLUSL(J+1)*UMINUS2L(J) -
            4*(J-LMIN+0.5)*(J+LMIN+0.5))**2.)


        _P12 = lambda J: \
            (((J-LMIN-1.5)*(J-LMIN-0.5))/
            (8*J*CMINUSL(J-1)*CPLUS2L(J)))*\
            ((UMINUSL(J-1)*UPLUS2L(J) - 4*(J-LMIN+0.5)*(J+LMIN+0.5))**2.)


        _Q12 = lambda J: \
            (((J-LMIN-0.5)*(J+0.5)*(J+LMIN+1.5))/(4*J*(J+1)*
            CMINUSL(J)*CPLUS2L(J)))*((UMINUSL(J)*UPLUS2L(J) -
            4*(J-LMIN+0.5)*(J+LMIN+0.5))**2.)


        _R12 = lambda J: \
            (((J+LMIN+1.5)*(J+LMIN+2.5))/(8*(J+1)*CMINUSL(J+1)*
            CPLUS2L(J)))*((UMINUSL(J+1)*UPLUS2L(J) -
            4*(J-LMIN+0.5)*(J+LMIN+0.5))**2.)


        _P2 = lambda J: \
            (((J-LMIN-1.5)*(J-LMIN-0.5))/(8*J*CPLUSL(J-1)*
            CPLUS2L(J)))*((UPLUSL(J-1)*UPLUS2L(J) +
            4*(J-LMIN+0.5)*(J+LMIN+0.5))**2.)


        _Q2 = lambda J: \
            (((J-LMIN-0.5)*(J+0.5)*(J+LMIN+1.5))/(4*J*(J+1)*
            CPLUSL(J)*CPLUS2L(J)))*((UPLUSL(J)*UPLUS2L(J) +
            4*(J-LMIN+0.5)*(J+LMIN+0.5))**2.)


        _R2 = lambda J: \
            (((J+LMIN+1.5)*(J+LMIN+2.5))/(8*(J+1)*CPLUSL(J+1)*
            CPLUS2L(J)))*((UPLUSL(J+1)*UPLUS2L(J) +
            4*(J-LMIN+0.5)*(J+LMIN+0.5))**2.)


        # Resolves the Delta Lambda = LAML - LAM2L

        if LAML > LAM2L:
            return (
             ("P1", _P1),
             ("Q1", _Q1),
             ("R1", _R1),
             ("P21", _P21),
             ("Q21", _Q21),
             ("R21", _R21),
             ("P12", _P12),
             ("Q12", _Q12),
             ("R12", _R12),
             ("P2", _P2),
             ("Q2", _Q2),
             ("R2", _R2),
            )

        elif LAML < LAM2L:
            return (
             ("P1", lambda J: _R1(J-1)),
             ("Q1", _Q1),
             ("R1", lambda J: _P1(J+1)),
             ("P21", lambda J: _R12(J-1)),
             ("Q21", _Q12),
             ("R21", lambda J: _P12(J+1)),
             ("P12", lambda J: _R21(J-1)),
             ("Q12", _Q21),
             ("R12", lambda J: _P21(J+1)),
             ("P2", lambda J: _R2(J-1)),
             ("Q2", _Q2),
             ("R2", lambda J: _P2(J+1)),
            )


class _LSTTriplet(_LineStrengthToolbox):
    """
    Base class for Triplet cases (Delta Lambda = [0, -1, +1]), containing "U" and "C" formulas

    Formulas: 2.1.4-8 to 2.1.4-10 (p69-p70)

    Adapted from Fortran code: ATMOS/wrk4/bruno/Mole/NH/sjnh.f
    """

    multiplicityl = 3
    multiplicity2l = 3
    quanta_to_branch = staticmethod(_quanta_to_branch_same_multiplicity)

    def _get_strengths(self, J, LAML, LAM2L, U1PLUSL, U1MINUSL, U3PLUSL, U3MINUSL,  C1L, C2L, C3L,
                       U1PLUS2L, U1MINUS2L, U3PLUS2L, U3MINUS2L, C12L, C22L, C32L, YL, Y2L):
        raise NotImplementedError()

    def _get_populate_data(self, vl, v2l, J):
        cc = self._molconsts
        LAML = cc["from_spdf"]
        LAM2L = cc["to_spdf"]
        AL = cc["statel_A"]
        AEL = cc["statel_alpha_e"]
        BEL = cc["statel_B_e"]
        A2L = cc["state2l_A"]
        AE2L = cc["state2l_alpha_e"]
        BE2L = cc["state2l_B_e"]

        BL = BEL - AEL * (vl + 0.5)
        B2L = BE2L - AE2L * (v2l + 0.5)

        YL = AL / BL
        Y2L = A2L / B2L

        # Lambdas for formulas C1 and C3 (see "p131" comment below)
        LAMLC = LAML
        LAM2LC = LAM2L

        # p61: “[...] Y= A/B. The term is called normal if A > 0, and inverted if A < 0”
        # p131: “in the case of inverted terms, Y is to be substituted with negative sign;
        #        furthermore, -Lambda is to be written instead of Lambda in C1(J) and C3(J)

        if YL < 0:
            YL = -YL
            LAMLC = -LAMLC
        if Y2L < 0:
            Y2L = -Y2L
            LAM2LC = -LAM2LC

        # Original comment:
        #
        #     cálculo das contantes U1+, U1-, U3+, U3-,
        #     C1+, C2+, C3+, dependentes de J para o estado
        #     eletronico superior( A 3PI, lambda=1)
        #     formulas para casos intermediarios entre os casos A( Y>>J(J+1) )
        #     e B( Y<<J(J+1) ) de Hund
        #
        # Note: I (JT) changed some entity names for clarity and generality. For example:
        #
        #     - U1PLUSL was originally called U1MAA, ("MA" stood for "MAIS", "plus" in Portuguese;
        #       "A" stood the particular state for which the original NH code was developed and was replaced
        #       by "L", meaning "linha", to be consistent with a notation found in many places in the code)
        #     - "YL" was replaced by "YL", and "Y2L" was replaced by "Y2L" because, again, "L" and "2L"
        #       are used in other places to denote the superior and inferior levels, respectively

        U1PLUSL = lambda J: math.sqrt(LAML*LAML*YL*(YL-4)+4*J*J)+(LAML*(YL-2))
        U1MINUSL = lambda J: math.sqrt(LAML*LAML*YL*(YL-4)+4*J*J)-(LAML*(YL-2))
        U3PLUSL = lambda J: (LAML*LAML*YL*(YL-4)+4*(J+1)*(J+1))**.5+LAML*(YL-2)
        U3MINUSL = lambda J: ((LAML*LAML*YL*(YL-4)+4*(J+1)*(J+1))**0.5)-LAML*(YL-2)
        C1L = lambda J: LAMLC*LAMLC*YL*(YL-4)*(J-LAMLC+1)*(J+LAMLC)+2*(2*J+1)*(J-LAMLC)*(J+LAMLC)*J
        C2L = lambda J: LAML*LAML*YL*(YL-4)+4*J*(J+1)
        C3L = lambda J: LAMLC*LAMLC*YL*(YL-4)*(J-LAMLC)*(J+LAMLC+1)+2*(2*J+1)*(J-LAMLC+1)*(J+1)*(J+LAMLC+1)

        U1PLUS2L = lambda J: ((LAM2L*LAM2L*Y2L*(Y2L-4)+4*J*J))**.5+LAM2L*(Y2L-2)
        U1MINUS2L = lambda J: ((LAM2L*LAM2L*Y2L*(Y2L-4)+4*J*J))**.5-LAM2L*(Y2L-2)
        U3PLUS2L = lambda J: ((LAM2L*LAM2L*Y2L*(Y2L-4)+4*(J+1)*(J+1))**.5)+LAM2L*(Y2L-2)
        U3MINUS2L = lambda J: ((LAM2L*LAM2L*Y2L*(Y2L-4)+4*(J+1)*(J+1))**.5)-LAM2L*(Y2L-2)
        C12L = lambda J: LAM2L*LAM2L*Y2L*(Y2L-4)*(J-LAM2L+1)*(J+LAM2L)+2*(2*J+1)*(J-LAM2L)*J*(J+LAM2L)
        C22L = lambda J: LAM2L*LAM2L*Y2L*(Y2L-4)+4*J*(J+1)
        C32L = lambda J: LAM2LC*LAM2LC*Y2L*(Y2L-4)*(J-LAM2LC)*(J+LAM2LC+1)+2*(2*J+1)*(J-LAM2LC+1)*(J+1)*(J+LAM2LC+1)

        ret = self._get_strengths(J, LAML, LAM2L,
                                  U1PLUSL, U1MINUSL, U3PLUSL, U3MINUSL,
                                  C1L, C2L, C3L,
                                  U1PLUS2L, U1MINUS2L, U3PLUS2L, U3MINUS2L,
                                  C12L, C22L, C32L, YL, Y2L)

        return ret


class _LSTTriplet0(_LSTTriplet):
    """
    Honl-London factors for (triplet, Delta Lambda = +-1)

    Formulas: Table 3.10 (p136-p137); Formulas 2.1.4-8 to 2.1.4-10 (p69-p70)

    Adapted from Fortran code: ATMOS/wrk4/bruno/Mole/NH/sjc2kovacs.for (C2 Swan)
    """

    absDeltaLambda = 0

    def _get_strengths(self, J, LAML, LAM2L, U1PLUSL, U1MINUSL, U3PLUSL, U3MINUSL,  C1L, C2L, C3L,
                       U1PLUS2L, U1MINUS2L, U3PLUS2L, U3MINUS2L, C12L, C22L, C32L, YL, Y2L):

        LAM = LAML  # LAML and LAM2L are the same


        _P1 = lambda J: \
         (((J-LAM)*(J+LAM))/(16*J*C1L(J-1)*C12L(J)))* \
         (((J-LAM+1)*(J+LAM-1)*U1PLUSL(J-1)*U1PLUS2L(J)+
         (J-LAM-1)*(J+LAM+1)*U1MINUSL(J-1)*U1MINUS2L(J)+
         8*(J-LAM-1)*(J-LAM)*(J+LAM-1)*(J+LAM))**2)

        _Q1 = lambda J: \
         ((2*J+1)/(16*J*(J+1)*C1L(J)*C12L(J))) \
         *(((LAM-1)*(J-LAM+1)*(J+LAM)*U1PLUSL(J)*U1PLUS2L(J)+
         (LAM+1)*(J-LAM)*(J+LAM+1)*U1MINUSL(J)*U1MINUS2L(J)+
         8*LAM*(J-LAM)*(J-LAM)*(J+LAM)*(J+LAM))**2)

        _R1 = lambda J: \
         (((J-LAM+1)*(J+LAM+1))/(16*(J+1)*
         C1L(J+1)*C12L(J)))*(((J-LAM+2)*(J+LAM)*
         U1PLUSL(J+1)*U1PLUS2L(J)+(J-LAM)*(J+LAM+2)*U1MINUSL(J+1)*
         U1MINUS2L(J)+8*(J-LAM)*(J-LAM+1)*(J+LAM)*(J+LAM+1))**2)

        _P21 = lambda J: \
         (((J-LAM)*(J+LAM))/(2*J*C2L(J-1)*C12L(J)))* \
         (((J-LAM+1)*(J+LAM-1)*U1PLUS2L(J)-
         (J-LAM-1)*(J+LAM+1)*U1MINUS2L(J)-
         2*LAM*(J-LAM)*(J+LAM)*(YL-2))**2)

        _Q21 = lambda J: \
         ((2*J+1)/(2*J*(J+1)*C2L(J)*C12L(J))) \
         *(((LAM-1)*(J-LAM+1)*(J+LAM)*U1PLUS2L(J)-
         (LAM+1)*(J-LAM)*(J+LAM+1)*U1MINUS2L(J)-
         2*LAM*LAM*(J-LAM)*(J+LAM)*(YL-2))**2)

        _R21 = lambda J: \
         (((J-LAM+1)*(J+LAM+1))/(2*(J+1)*
         C2L(J+1)*C12L(J)))*(((J-LAM+2)*(J+LAM)*
         U1PLUS2L(J)-(J-LAM)*(J+LAM+2)*U1MINUS2L(J)-
         2*LAM*(J-LAM)*(J+LAM)*(YL-2))**2)

        _P31 = lambda J: \
         (((J-LAM)*(J+LAM))/(16*J*C3L(J-1)*C12L(J)))* \
         (((J-LAM+1)*(J+LAM-1)*U3MINUSL(J-1)*U1PLUS2L(J)+
         (J-LAM-1)*(J+LAM+1)*U1MINUS2L(J)*U3PLUSL(J-1)-
         8*(J-LAM)*(J+LAM)*(J-LAM)*(J+LAM))**2)

        _Q31 = lambda J: \
         ((2*J+1)/(16*J*(J+1)*C32L(J)*C12L(J))) \
         *(((LAM-1)*(J-LAM+1)*(J+LAM)*U3MINUSL(J)*U1PLUS2L(J)+
         (LAM+1)*(J-LAM)*(J+LAM+1)*U3PLUSL(J)*U1MINUS2L(J)-
         8*LAM*(J-LAM)*(J-LAM+1)*(J+LAM)*(J+LAM+1))**2)

        _R31 = lambda J: \
         (((J-LAM+1)*(J+LAM+1))/(16*(J+1)*
         C3L(J+1)*C12L(J)))*(((J-LAM+2)*(J+LAM)*U3MINUSL(J+1)*
         U1PLUS2L(J)+(J-LAM)*(J+LAM+2)*U3PLUSL(J+1)*U1MINUS2L(J)-
         8*(J-LAM)*(J-LAM+2)*(J+LAM)*(J+LAM+2))**2)

        _P12 = lambda J: \
         (((J-LAM)*(J+LAM))/(2*J*C1L(J-1)*C22L(J)))* \
         (((J-LAM+1)*(J+LAM-1)*U1PLUSL(J-1)-
         (J-LAM-1)*(J+LAM+1)*U1MINUSL(J-1)-
         2*LAM*(J-LAM-1)*(J+LAM-1)*(Y2L-2))**2)

        _Q12 = lambda J: \
         ((2*J+1)/(2*J*(J+1)*C1L(J)*C22L(J))) \
         *(((LAM-1)*(J-LAM+1)*(J+LAM)*U1PLUSL(J)-
         (LAM+1)*(J-LAM)*(J+LAM+1)*U1MINUSL(J)-
         2*LAM*LAM*(J-LAM)*(J+LAM)*(Y2L-2))**2)

        _R12 = lambda J: \
         (((J-LAM+1)*(J+LAM+1))/(2*(J+1)*
         C1L(J+1)*C22L(J)))*(((J-LAM+2)*(J+LAM)*
         U1PLUSL(J+1)-(J-LAM)*(J+LAM+2)*U1MINUSL(J+1)-
         2*LAM*(J-LAM+1)*(J+LAM+1)*(Y2L-2))**2)

        _P2 = lambda J: \
         ((4*(J-LAM)*(J+LAM))/(J*C2L(J-1)*C22L(J)))* \
         ((.5*LAM*LAM*(YL-2)*(Y2L-2)+(J-LAM-1)*(J+LAM+1)+
         (J-LAM+1)*(J+LAM-1))**2)

        _Q2 = lambda J: \
         ((4*(2*J+1))/(J*(J+1)*C2L(J)*C22L(J)))* \
         ((.5*LAM*LAM*LAM*(YL-2)*(Y2L-2)+
         (LAM+1)*(J-LAM)*(J+LAM+1)+(LAM-1)*(J-LAM+1)*
         (J+LAM))**2)

        _R2 = lambda J: \
         ((4*(J-LAM+1)*(J+LAM+1))/((J+1)*
         C2L(J+1)*C22L(J)))*((.5*LAM*LAM*(YL-2)*(Y2L-2)+
         (J-LAM)*(J+LAM+2)+(J-LAM+2)*(J+LAM))**2)

        _P32 = lambda J: \
         (((J-LAM)*(J+LAM))/(2*J*C3L(J-1)*C22L(J)))* \
         (((J-LAM+1)*(J+LAM-1)*U3MINUSL(J-1)-
         (J-LAM-1)*(J+LAM+1)*U3PLUSL(J-1)+
         2*LAM*(J-LAM)*(J+LAM)*(Y2L-2))**2)

        _Q32 = lambda J: \
         ((2*J+1)/(2*J*(J+1)*C3L(J)*C22L(J))) \
         *(((LAM-1)*(J-LAM+1)*(J+LAM)*U3MINUSL(J)-
         (LAM+1)*(J-LAM)*(J+LAM+1)*U3PLUSL(J)+
         2*LAM*LAM*(J-LAM+1)*(J+LAM+1)*(Y2L-2))**2)

        _R32 = lambda J: \
         (((J-LAM+1)*(J+LAM+1))/(2*(J+1)*
         C3L(J+1)*C22L(J)))*(((J-LAM+2)*(J+LAM)*
         U3MINUSL(J+1)-(J-LAM)*(J+LAM+2)*U3PLUSL(J+1)+
         2*LAM*(J-LAM+2)*(J+LAM+2)*(Y2L-2))**2)

        _P13 = lambda J: \
         (((J-LAM)*(J+LAM))/(16*J*C1L(J-1)*C32L(J)))* \
         (((J-LAM+1)*(J+LAM-1)*U1PLUSL(J)*U3MINUS2L(J)+
         (J-LAM-1)*(J+LAM+1)*U1MINUSL(J-1)*U3PLUS2L(J)-
         8*(J-LAM-1)*(J-LAM+1)*(J+LAM-1)*(J+LAM+1))**2)

        _Q13 = lambda J: \
         ((2*J+1)/(16*J*(J+1)*C1L(J)*C32L(J))) \
         *(((LAM-1)*(J-LAM+1)*(J+LAM)*U1PLUSL(J)*U3MINUS2L(J)+
         (LAM+1)*(J-LAM)*(J+LAM+1)*U1MINUSL(J)*U3PLUS2L(J)-
         8*LAM*(J-LAM)*(J-LAM+1)*(J+LAM)*(J+LAM+1))**2)

        _R13 = lambda J: \
         (((J-LAM+1)*(J+LAM+1))/(16*(J+1)*
         C1L(J+1)*C32L(J)))*(((J-LAM+2)*(J+LAM)*
         U1PLUSL(J+1)*U3MINUS2L(J)+(J-LAM)*(J+LAM+2)*U1MINUSL(J+1)*
         U3PLUS2L(J)-8*(J-LAM+1)*(J-LAM+1)*(J+LAM+1)*(J+LAM+1))**2)

        _P23 = lambda J: \
          (((J-LAM)*(J+LAM))/(2*J*C2L(J-1)*C32L(J)))* \
          (((J-LAM+1)*(J+LAM-1)*U3MINUS2L(J)-
          (J-LAM-1)*(J+LAM+1)*U3PLUS2L(J)+
          2*LAM*(J-LAM+1)*(J+LAM+1)*(YL-2))**2)

        _Q23 = lambda J: \
         ((2*J+1)/(2*J*(J+1)*C2L(J)*C32L(J))) \
         *(((LAM-1)*(J-LAM+1)*(J+LAM)*U3MINUS2L(J)-
         (LAM+1)*(J-LAM)*(J+LAM+1)*U3PLUS2L(J)+
         2*LAM*LAM*(J-LAM+1)*(J+LAM+1)*(YL-2))**2)

        _R23 = lambda J: \
         (((J-LAM+1)*(J+LAM+1))/(2*(J+1)*
         C2L(J+1)*C32L(J)))*(((J-LAM+2)*(J+LAM)*
         U3MINUS2L(J)-(J-LAM)*(J+LAM+2)*U3PLUS2L(J)+
         2*LAM*(J-LAM+1)*(J+LAM+1)*(YL-2))**2)

        _P3 = lambda J: \
         (((J-LAM)*(J+LAM))/(16*J*C3L(J-1)*
         C32L(J)))*(((J-LAM+1)*(J+LAM-1)*U3MINUSL(J-1)*
         U3MINUS2L(J)+(J-LAM-1)*(J+LAM+1)*U3PLUSL(J-1)*U3PLUS2L(J)+
         8*(J-LAM)*(J-LAM+1)*(J+LAM)*(J+LAM+1))**2)

        _Q3 = lambda J: \
         ((2*J+1)/(16*J*(J+1)*C3L(J)*C32L(J)))* \
         (((LAM-1)*(J-LAM+1)*(J+LAM)*U3MINUSL(J)*U3MINUS2L(J)+
         (LAM+1)*(J-LAM)*(J+LAM+1)*U3PLUSL(J)*U3PLUS2L(J)+
         8*LAM*(J-LAM+1)*(J-LAM+1)*(J+LAM+1)*(J+LAM+1))**2)

        _R3 = lambda J: \
         (((J-LAM+1)*(J+LAM+1))/(16*(J+1)*
         C3L(J+1)*C32L(J)))*(((J-LAM+2)*(J+LAM)*U3MINUSL(J+1)
         *U3MINUS2L(J)+(J-LAM)*(J+LAM+2)*U3PLUSL(J+1)*U3PLUS2L(J)+
         8*(J-LAM+1)*(J-LAM+2)*(J+LAM+1)*(J+LAM+2))**2)

        return (
             ("P1", _P1),
             ("Q1", _Q1),
             ("R1", _R1),
             ("P21", _P21),
             ("Q21", _Q21),
             ("R21", _R21),
             ("P31", _P31),
             ("Q31", _Q31),
             ("R31", _R31),
             ("P12", _P12),
             ("Q12", _Q12),
             ("R12", _R12),
             ("P2", _P2),
             ("Q2", _Q2),
             ("R2", _R2),
             ("P32", _P32),
             ("Q32", _Q32),
             ("R32", _R32),
             ("P13", _P13),
             ("Q13", _Q13),
             ("R13", _R13),
             ("P23", _P23),
             ("Q23", _Q23),
             ("R23", _R23),
             ("P3", _P3),
             ("Q3", _Q3),
             ("R3", _R3),
            )


class _LSTTriplet1(_LSTTriplet):
    """
    Honl-London factors for (triplet, Delta Lambda = +-1)

    Formulas: Table 3.10 (p136-p137); Formulas 2.1.4-8 to 2.1.4-10 (p69-p70)

    Adapted from Fortran code: ATMOS/wrk4/bruno/Mole/NH/sjnh.f
    """

    absDeltaLambda = 1

    def _get_strengths(self, J, LAML, LAM2L, U1PLUSL, U1MINUSL, U3PLUSL, U3MINUSL,  C1L, C2L, C3L,
                       U1PLUS2L, U1MINUS2L, U3PLUS2L, U3MINUS2L, C12L, C22L, C32L, YL, Y2L):

        # Original comment:
        #
        #     usa-se o menor valor de Lambda nas formulas abaixo
        #
        # p139: "The Lambda's occurring explicitly in the formulas mean once more the smaller
        #        of the two Lambda's involved in the transition"
        LMIN = min(LAML, LAM2L)

        _P1 = lambda J: \
            ((J-LMIN-1)*(J-LMIN))/(32*J*C1L(J-1)*C12L(J))* \
            (((J-LMIN+1)*(J+LMIN)*U1PLUSL(J-1)*U1PLUS2L(J)+
            (J-LMIN-2)*(J+LMIN+1)*U1MINUSL(J-1)*U1MINUS2L(J)+
            8*(J-LMIN-2)*(J-LMIN)*(J+LMIN)*(J+LMIN))**2)

        _Q1 = lambda J: \
            ((J-LMIN)*(J+LMIN+1)*(2*J+1))/(32*J*(J+1)*C1L(J)*C12L(J)) \
            *(((J-LMIN+1)*(J+LMIN)*U1PLUSL(J)*U1PLUS2L(J)+
            (J-LMIN-1)*(J+LMIN+2)*U1MINUSL(J)*U1MINUS2L(J)+
            8*(J-LMIN-1)*(J-LMIN)*(J+LMIN)*(J+LMIN+1))**2)

        _R1 = lambda J: \
            (((J+LMIN+1)*(J+LMIN+2))/(32*(J+1)*
            C1L(J+1)*C12L(J)))*(((J-LMIN+1)*(J+LMIN)*
            U1PLUSL(J+1)*U1PLUS2L(J)+(J-LMIN)*(J+LMIN+3)*U1MINUSL(J+1)*
            U1MINUS2L(J)+8*(J-LMIN)*(J-LMIN)*(J+LMIN)*(J+LMIN+2))**2)


        _P21 = lambda J: \
            (((J-LMIN-1)*(J-LMIN))/(4*J*C2L(J-1)*C12L(J)))* \
            (((J-LMIN+1)*(J+LMIN)*U1PLUS2L(J)-
            (J-LMIN-2)*(J+LMIN+1)*U1MINUS2L(J)-
            2*(LMIN+1)*(J-LMIN)*(J+LMIN)*(YL-2))**2)

        _Q21 = lambda J: \
            (((J-LMIN)*(J+LMIN+1)*(2*J+1))/(4*J*(J+1)*C2L(J)*C12L(J)))* \
            (((J-LMIN+1)*(J+LMIN)*U1PLUS2L(J)-
            (J-LMIN-1)*(J+LMIN+2)*U1MINUS2L(J)-
            2*(LMIN+1)*(J-LMIN)*(J+LMIN)*(YL-2))**2)

        _R21 = lambda J: \
            (((J+LMIN+1)*(J+LMIN+2))/(4*(J+1)*
            C2L(J+1)*C12L(J)))*(((J-LMIN+1)*(J+LMIN)*
            U1PLUS2L(J) - (J-LMIN)*(J+LMIN+3)*U1MINUS2L(J)
            - 2*(LMIN+1)*(J-LMIN)*(J+LMIN)*(YL-2))**2)

        _P31 = lambda J: \
            (((J-LMIN-1)*(J-LMIN))/(32*J*C3L(J-1)*C12L(J)))* \
            (((J-LMIN+1)*(J+LMIN)*U3MINUSL(J-1)*U1PLUS2L(J)+
            (J-LMIN-2)*(J+LMIN+1)*U3PLUSL(J-1)*U1MINUS2L(J)-
            8*(J-LMIN-1)*(J-LMIN)*(J+LMIN)*(J+LMIN+1))**2)

        _Q31 = lambda J: \
            (((J-LMIN)*(J+LMIN+1)*(2*J+1))/(32*J*(J+1)*C3L(J)*C12L(J))) \
            *(((J-LMIN+1)*(J+LMIN)*U3MINUSL(J)*U1PLUS2L(J)+
            (J-LMIN-1)*(J+LMIN+2)*U3PLUSL(J)*U1MINUS2L(J)-
            8*(J-LMIN)*(J-LMIN)*(J+LMIN)*(J+LMIN+2))**2)

        _R31 = lambda J: \
            (((J+LMIN+1)*(J+LMIN+2))/(32*(J+1)*
            C3L(J+1)*C12L(J)))*(((J-LMIN+1)*(J+LMIN)*
            U3MINUSL(J+1)*U1PLUS2L(J)+(J-LMIN)*(J+LMIN+3)*U3PLUSL(J+1)*
            U1MINUS2L(J)- 8*(J-LMIN)*(J-LMIN+1)*(J+LMIN)*(J+LMIN+3))**2)

        _P12 = lambda J: \
            (((J-LMIN-1)*(J-LMIN))/(4*J*C1L(J-1)*C22L(J)))* \
            (((J-LMIN+1)*(J+LMIN)*U1PLUSL(J-1)-
            (J-LMIN-2)*(J+LMIN+1)*U1MINUSL(J-1)-
            2*LMIN*(J-LMIN-2)*(J+LMIN)*(Y2L-2))**2)

        _Q12 = lambda J: \
            (((J-LMIN)*(J+LMIN+1)*(2*J+1))/(4*J*(J+1)*C1L(J)*C22L(J))) \
            *(((J-LMIN+1)*(J+LMIN)*U1PLUSL(J)-
            (J-LMIN-1)*(J+LMIN+2)*U1MINUSL(J)-
            2*LMIN*(J-LMIN-1)*(J+LMIN+1)*(Y2L-2))**2)

        _R12 = lambda J: \
            (((J+LMIN+1)*(J+LMIN+2))/(4*(J+1)*
            C1L(J+1)*C22L(J)))*(((J-LMIN+1)*(J+LMIN)*
            U1PLUSL(J+1) - (J-LMIN)*(J+LMIN+3)*U1MINUSL(J+1)
            - 2*LMIN*(J-LMIN)*(J+LMIN+2)*(Y2L-2))**2)

        _P2 = lambda J: \
            ((2*(J-LMIN-1)*(J-LMIN))/(J*C2L(J-1)*C22L(J)))* \
            ((.5*LMIN*(LMIN+1)*(YL-2)*(Y2L-2)+(J-LMIN+1)*(J+LMIN)+
            (J-LMIN-2)*(J+LMIN+1))**2)

        _Q2 = lambda J: \
            ((2*(J-LMIN)*(J+LMIN+1)*(2*J+1))/(J*(J+1)*C2L(J)*C22L(J)))* \
            ((.5*LMIN*(LMIN+1)*(YL-2)*(Y2L-2)+
            (J-LMIN+1)*(J+LMIN)+(J-LMIN-1)*(J+LMIN+2))**2)

        _R2 = lambda J: \
            ((2*(J+LMIN+1)*(J+LMIN+2))/((J+1)*
            C2L(J+1)*C22L(J)))*((.5*LMIN*(LMIN+1)*(YL-2)*(Y2L-2)+
            (J-LMIN+1)*(J+LMIN)+(J-LMIN)*(J+LMIN+3))**2)

        _P32 = lambda J: \
            (((J-LMIN-1)*(J-LMIN))/(4*J*C3L(J-1)*C22L(J)))* \
            (((J-LMIN+1)*(J+LMIN)*U3MINUSL(J-1)-
            (J-LMIN-2)*(J+LMIN+1)*U3PLUSL(J-1)+
            2*LMIN*(J-LMIN-1)*(J+LMIN+1)*(Y2L-2))**2)

        _Q32 = lambda J: \
            (((J-LMIN)*(J+LMIN+1)*(2*J+1))/(4*J*(J+1)*C3L(J)*C22L(J))) \
            *(((J-LMIN+1)*(J+LMIN)*U3MINUSL(J)-
            (J-LMIN-1)*(J+LMIN+2)*U3PLUSL(J)+
            2*LMIN*(J-LMIN)*(J+LMIN+2)*(Y2L-2))**2)

        _R32 = lambda J: \
            (((J+LMIN+1)*(J+LMIN+2))/(4*(J+1)*
            C3L(J+1)*C22L(J)))*(((J-LMIN+1)*(J+LMIN)*
            U3MINUSL(J+1) - (J-LMIN)*(J+LMIN+3)*U3PLUSL(J+1)
            + 2*LMIN*(J-LMIN+1)*(J+LMIN+3)*(Y2L-2))**2)

        _P13 = lambda J: \
            (((J-LMIN-1)*(J-LMIN))/(32*J*C1L(J-1)*
            C32L(J)))*(((J-LMIN+1)*(J+LMIN)*U1PLUSL(J-1)*
            U3MINUS2L(J)+ (J-LMIN-2)*(J+LMIN+1)*U1MINUSL(J-1)*U3PLUS2L(J)-
            8*(J-LMIN-2)*(J-LMIN+1)*(J+LMIN)*(J+LMIN+1))**2)

        _Q13 = lambda J: \
            (((J-LMIN)*(J+LMIN+1)*(2*J+1))/(32*J*(J+1)*C1L(J)*C32L(J)))* \
            (((J-LMIN+1)*(J+LMIN)*U1PLUSL(J)*U3MINUS2L(J)+
            (J-LMIN-1)*(J+LMIN+2)*U1MINUSL(J)*U3PLUS2L(J)-
            8*(J-LMIN-1)*(J-LMIN+1)*(J+LMIN+1)*(J+LMIN+1))**2)

        _R13 = lambda J: \
            (((J+LMIN+1)*(J+LMIN+2))/(32*(J+1)*
            C1L(J+1)*C32L(J)))*(((J-LMIN+1)*(J+LMIN)*U1PLUSL(J+1)
            *U3MINUS2L(J)+(J-LMIN)*(J+LMIN+3)*U1MINUSL(J+1)*U3PLUS2L(J)-
            8*(J-LMIN)*(J-LMIN+1)*(J+LMIN+1)*(J+LMIN+2))**2)

        _P23 = lambda J: \
            (((J-LMIN-1)*(J-LMIN))/(4*J*C2L(J-1)*C32L(J)))* \
            (((J-LMIN+1)*(J+LMIN)*U3MINUS2L(J)-
            (J-LMIN-2)*(J+LMIN+1)*U3PLUS2L(J)+
            2*(LMIN+1)*(J-LMIN+1)*(J+LMIN+1)*(YL-2))**2)

        _Q23 = lambda J: \
            (((J-LMIN)*(J+LMIN+1)*(2*J+1))/(4*J*(J+1)*C2L(J)*C32L(J))) \
            *(((J-LMIN+1)*(J+LMIN)*U3MINUS2L(J)-
            (J-LMIN-1)*(J+LMIN+2)*U3PLUS2L(J)+
            2*(LMIN+1)*(J-LMIN+1)*(J+LMIN+1)*(YL-2))**2)

        _R23 = lambda J: \
            (((J+LMIN+1)*(J+LMIN+2))/(4*(J+1)*
            C2L(J+1)*C32L(J)))*(((J-LMIN+1)*(J+LMIN)*
            U3MINUS2L(J) - (J-LMIN)*(J+LMIN+3)*U3PLUS2L(J)
            + 2*(LMIN+1)*(J-LMIN+1)*(J+LMIN+1)*(YL-2))**2)

        _P3 = lambda J: \
            (((J-LMIN-1)*(J-LMIN))/(32*J*C3L(J-1)*
            C32L(J)))*(((J-LMIN+1)*(J+LMIN)*U3MINUSL(J-1)*
            U3MINUS2L(J)+(J-LMIN-2)*(J+LMIN+1)*U3PLUSL(J-1)*U3PLUS2L(J)+
            8*(J-LMIN-1)*(J-LMIN+1)*(J+LMIN+1)*(J+LMIN+1))**2)

        _Q3 = lambda J: \
            (((J-LMIN)*(J+LMIN+1)*(2*J+1))/(32*J*(J+1)*C3L(J)*C32L(J)))* \
            (((J-LMIN+1)*(J+LMIN)*U3MINUSL(J)*U3MINUS2L(J)+
            (J-LMIN-1)*(J+LMIN+2)*U3PLUSL(J)*U3PLUS2L(J)+
            8*(J-LMIN)*(J-LMIN+1)*(J+LMIN+1)*(J+LMIN+2))**2)

        _R3 = lambda J: \
            (((J+LMIN+1)*(J+LMIN+2))/(32*(J+1)*
            C3L(J+1)*C32L(J)))*(((J-LMIN+1)*(J+LMIN)*U3MINUSL(J+1)
            *U3MINUS2L(J)+(J-LMIN)*(J+LMIN+3)*U3PLUSL(J+1)*U3PLUS2L(J)+
            8*(J-LMIN+1)*(J-LMIN+1)*(J+LMIN+1)*(J+LMIN+3))**2)

        # Resolves the Delta Lambda

        if LAML > LAM2L:
            return (
             ("P1", _P1),
             ("Q1", _Q1),
             ("R1", _R1),
             ("P21", _P21),
             ("Q21", _Q21),
             ("R21", _R21),
             ("P31", _P31),
             ("Q31", _Q31),
             ("R31", _R31),
             ("P12", _P12),
             ("Q12", _Q12),
             ("R12", _R12),
             ("P2", _P2),
             ("Q2", _Q2),
             ("R2", _R2),
             ("P32", _P32),
             ("Q32", _Q32),
             ("R32", _R32),
             ("P13", _P13),
             ("Q13", _Q13),
             ("R13", _R13),
             ("P23", _P23),
             ("Q23", _Q23),
             ("R23", _R23),
             ("P3", _P3),
             ("Q3", _Q3),
             ("R3", _R3),
            )

        elif LAML < LAM2L:
            return (
             ("P1", lambda J: _R1(J-1)),
             ("Q1", _Q1),
             ("R1", lambda J: _P1(J+1)),
             ("P21", _R12(J-1)),
             ("Q21", _Q12),
             ("R21", lambda J: _P12(J+1)),
             ("P31", lambda J: _R13(J-1)),
             ("Q31", _Q13),
             ("R31", lambda J: _P13(J+1)),
             ("P12", lambda J: _R21(J-1)),
             ("Q12", _Q21),
             ("R12", lambda J: _P21(J+1)),
             ("P2", lambda J: _R2(J-1)),
             ("Q2", _Q2),
             ("R2", lambda J: _P2(J+1)),
             ("P32", lambda J: _R23(J-1)),
             ("Q32", _Q23),
             ("R32", lambda J: _P23(J+1)),
             ("P13", lambda J: _R31(J-1)),
             ("Q13", _Q31),
             ("R13", lambda J: _P31(J+1)),
             ("P23", lambda J: _R32(J-1)),
             ("Q23", _Q32),
             ("R23", lambda J: _P32(J+1)),
             ("P3", lambda J: _R3(J-1)),
             ("Q3", _Q3),
             ("R3", lambda J: _P3(J+1)),
            )




"""FileMAbFwhm class (differential ABundances and FWHMs)"""


__all__ = ["FileAbXFwhm"]


from f311 import FilePy
from ..basic import adjust_atomic_symbol
import imp
import numpy as np
import a99


_COMMENT0 = """# Specification of differential abundances for each chemical.
# Differential abundance is a number to add to original value in abonds.dat.
# All lists of abundances for each element must have the same length."""
_COMMENT1 = """# Name for each pfant run. They will be part of spectrum file names.
# This is optional. If not used, sequential numbers will be used instead.
# Example: pfant_names = ["A", "B", "C", "D"]"""
_COMMENT2 = """# Convolutions specification for fwhm parameter:
# [first value, last value, step]"""


@a99.froze_it
class FileAbXFwhm(FilePy):
    __doc__ = """`x.py` Differential Abundances X FWHMs (Python source)

This file is actually Python source. Here is a sample:

%s
ab = {"Ca": [-.3, 0, .3, .5],
      "Si": [-.3, 0, .3, .5]}

%s
pfant_names = []

%s
conv = [0.08, 0.6,  0.04]
""" % (_COMMENT0, _COMMENT1, _COMMENT2)

    default_filename = "abxfwhm.py"
    attrs = ["ab", "conv"]
    editors = ["x.py"]

    @property
    def ab(self):
        """Abundances dictionary.

        Setting this property will cause source to be rebuilt and comments to
        be lost."""
        return self.__ab

    @ab.setter
    def ab(self, x):
        self.__ab = x
        self.__adjust_atomic_symbols()
        self.__rebuild_source()

    @property
    def conv(self):
        """Convolutions FWHM list.

        Setting this property will cause source to be rebuilt and comments to
        be lost."""
        return self.__conv

    @conv.setter
    def conv(self, x):
        self.__conv = x
        self.__rebuild_source()

    @property
    def pfant_names(self):
        """List of "names" for each pfant run.

        Must be empty or be of same size as any list of self.ab"""
        return self.__pfant_names

    @pfant_names.setter
    def pfant_names(self, x):
        self.__pfant_names = x


    @property
    def source(self):
        """Source code.

        Better to set this property than to set ab/conv. Setting this property
        will preserve the source code, whereas setting ab/conv separately will
        cause the source code to be rebuilt and become as boring as possible.
        """
        return self.__source

    @source.setter
    def source(self, x):
        self.__parse(x)

    def __init__(self):
        FilePy.__init__(self)

        self.__flag_parsed = False
        self.__flag_rebuild = False
        self.__source = ""
        self.__ab = None
        self.__conv = None
        self.__pfant_names = None

    def get_fwhms(self):
        """Returns FWHM's as a numpy vector."""
        return np.arange(self.conv[0], self.conv[1]+.00000001, self.conv[2])

    def validate(self, file_abonds=None):
        # validates abundances specification
        flag_first = True
        for symbol, mab in list(self.__ab.items()):
            assert isinstance(mab, (list, tuple)), \
                'Symbol "%s": differential abundances must be list or tuple' % symbol
            if flag_first:
                n = len(mab)
                flag_first = False
            else:
                if len(mab) != n:
                    raise ValueError('Symbol "%s": should have %d differential abundance%s, not %d' %
                     (symbol.strip(), n, "s" if n != 1 else "", len(mab)))
                # TODO: cross-check with ABONDS

        if self.__pfant_names:
            if not isinstance(self.__pfant_names, (list, tuple)):
                raise RuntimeError('"pfant_names" must be a list or tuple.')
            if len(self.__ab) > 0:
                for v in list(self.__ab.values()):
                    if len(v) != len(self.__pfant_names):
                        raise RuntimeError('"pfant_names" must be empty or have size %d.' % len(v))
                    break


        # validates if can use FWHM spect to make a vector
        try:
            fwhms = self.get_fwhms()
        except Exception as e:
            raise Exception('Error in "conv" specification: '+str(e))

        # this validation is necessary just because fwhm will be used as part of
        # filename
        for fwhm in self.get_fwhms():
            if fwhm > 9.99:
                raise RuntimeError("fhwm maximum is 9.99")


        if file_abonds:
            for symbol in self.ab:
                if not symbol in file_abonds.ele:
                    raise RuntimeError('Symbol "%s" is not in abundances list.' %
                                       symbol.strip())

    def __parse(self, x):
        """Populates __ab, __conf, and __source."""
        cfg = imp.new_module('cfg')
        exec(x, cfg.__dict__)
        self.__ab = cfg.ab
        self.__adjust_atomic_symbols()
        self.__conv = cfg.conv
        self.__pfant_names = None
        if hasattr(cfg, "pfant_names"):
            self.__pfant_names = cfg.pfant_names
        self.__source = x

    def __rebuild_source(self):
        self.__source = self._get_magic()+"\n%s\n%s\n\n%s\n%s\n" % (_COMMENT0,
         "ab = "+repr(self.__ab).replace("], ", "],\n      "),
         _COMMENT1, "conv = "+repr(self.__conv))

    def __adjust_atomic_symbols(self):
        """replaces self.__ab keys with the adjusted atomic symbols."""
        for key in self.__ab:
            self.__ab[adjust_atomic_symbol(key)] = self.__ab.pop(key)

    def _do_load(self, filename):
        with open(filename, "r") as h:
            self.__parse(h.read())

    def _do_save_as(self, filename):
        with open(filename, "w") as h:
            h.write(self.__source)




#!/usr/bin/env python
"""
Cuts atomic lines file to wavelength interval specified

The interval is [llzero, llfin]
"""

# TODO add this functionality to ated.py


import argparse
import logging
import a99
import pyfant


a99.logging_level = logging.INFO
a99.flag_log_file = True


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
     description=__doc__,
     formatter_class=a99.SmartFormatter
     )
    parser.add_argument('llzero', type=float, nargs=1,
     help='lower wavelength boundary (angstrom)')
    parser.add_argument('llfin', type=float, nargs=1,
     help='upper wavelength boundary (angstrom)')
    parser.add_argument('fn_input', type=str, help='input file name', nargs=1)
    parser.add_argument('fn_output', type=str, help='output file name', nargs=1)

    args = parser.parse_args()

    file_atoms = pyfant.FileAtoms()
    file_atoms.load(args.fn_input[0])

    m = min([min(a.lambda_) for a in file_atoms.atoms])
    M = max([max(a.lambda_) for a in file_atoms.atoms])

    print("Original interval: [%g, %g]" % (m, M))
    print("New interval: [%g, %g]" % (args.llzero[0], args.llfin[0]))

    file_atoms.cut(args.llzero[0], args.llfin[0])
    file_atoms.save_as(args.fn_output[0])
    print("Successfully created file '%s'" % args.fn_output[0])



__all__ = ["XFileOptions"]

from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import os.path
import a99
import f311
import pyfant


class XFileOptions(f311.XFileMainWindow):
    """Window to edit a FileOptions"""
    def _add_stuff(self):
        e0 = self.w = pyfant.WOptionsEditor(self)
        e0.changed.connect(self._on_w_changed)
        e0.loaded.connect(self._on_w_loaded)
        self.tabWidget.addTab(e0, "")
        # self.setCentralWidget(self.w)

        CLS_FILE = pyfant.FileOptions
        WILD = "*.py" if CLS_FILE.default_filename is None else "*"+os.path.splitext(CLS_FILE.default_filename)[1]
        self.pages.append(f311.MyPage(text_tab="File",
                                    cls_save=CLS_FILE, clss_load=(CLS_FILE,), wild=WILD, editor=e0))
        self.setWindowTitle("{} Editor".format(a99.get_obj_doc0(CLS_FILE)))

    def _on_w_loaded(self):
        pass

    def _on_w_changed(self):
        pass



"""
NIST Downloader
"""
from robobrowser import RoboBrowser
import bs4
import re
import unicodedata
import pyfant

__all__ = ["get_nist_webbook_constants", "NIST_URL"]


NIST_URL = "http://webbook.nist.gov/chemistry/form-ser.html"


def get_nist_webbook_constants(formula, flag_unicode=True, flag_parse_state=False):
    """
    Navigates through NIST webbook pages to retrieve a table of molecular constants

    Args:
        formula: example: "OH"
        flag_unicode:
        flag_parse_state: parses NIST 'State' column and extends each data row with three extra
                          columns: 'label', 'mult', 'spdf'

    Returns: tuple: table (list of lists), header (list of strings), name of molecule

    **Disclaimer** This scraper matches a specific version of the Chemistry Web Book. Therefore,
    it may stop working if the NIST web site is updated.
    """

    # Callables to be used in different conversion situations
    _conv_formula = lambda x: _html_formula_to_plain(x, flag_unicode)
    f_header = _conv_formula
    f_state = _conv_formula
    f_float_or_none = lambda x: _nonify(_floatify(_conv_formula(x)))
    f_str_or_none = lambda x: _nonify(_conv_formula(x))
    functions = [f_state] + [f_float_or_none] * 10 + [f_str_or_none, f_float_or_none]

    browser = RoboBrowser(history=True, parser="lxml")
    browser.open("http://webbook.nist.gov/chemistry/form-ser.html")

    # # Page
    form = browser.get_form()
    form["Formula"].value = formula
    browser.submit_form(form)

    # # Molecule park page
    text = browser.find(text="Constants of diatomic molecules")

    if text is None:
        # # Probably fell in this page showing possible choices
        # In this case, clicks on the first choice

        ol = browser.find("ol")

        if not ol:
            raise RuntimeError("Constants of diatomic molecules not found for '{}'".format(formula))

        link0 = ol.find("a")

        if not link0:
            raise RuntimeError("Constants of diatomic molecules not found for '{}'".format(formula))
        browser.follow_link(link0)

        # # Molecule park page
        text = browser.find(text="Constants of diatomic molecules")
        link_ctes = text.parent
        browser.follow_link(link_ctes)
    else:
        link_ctes = text.parent
        browser.follow_link(link_ctes)

    # # Page

    title = browser.find("title").text

    table = browser.find("table", class_="small data")

    rows = table.find_all("tr")
    n = 0
    header = None
    data = []  # list of lists
    for row in rows:
        cols = row.find_all("td")
        flag_header = False
        if len(cols) == 0:
            if n > 0:
                continue
            # If no columns found, we assume it is the header row
            cols = row.find_all("th")
            header = [f_header(ele) for ele in cols]
            header.append("A")

        elif len(cols) == 13:

            row = [function(ele) for function, ele in zip(functions, cols)]
            row.append(_parse_A(browser, cols[1]))
            data.append(row)
            n += 1

    if flag_parse_state:
        _extend_nist_table(data, header)

    return data, header, title


def _extend_nist_table(data, header):
    """Parses 'State' and adds 3 extra columns to table **in place**

    If parsing fails, (label, mult, spdf) columns for that row will be blank
    """

    header.extend(["label", "mult", "spdf"])

    for row in data:
        trio = ("", "", "")
        try:
            trio = _nist_str_to_state(row[0])
        except ValueError:
            pass
        row.extend(trio)


def _nist_str_to_state(s):
    """Parses NIST state string into tuple

    **Note** This method will deal with unicode superscript characters

    Args:
        s: NIST state string such as "E 1Sigma_g+" or "d ¹Sigma_u⁺"

    Return:
        (label, mult, spdf)
    """

    s_ = unicodedata.normalize('NFKD', s).encode('ascii', 'replace').decode()

    match = re.match("([A-Za-z])\s*(\d)\s*([A-Za-z]+)", s_)

    if match is None:
        raise ValueError("Nist 'State' string not recognized: {}".format(s))

    label, _mult, _spdf = match.groups()
    mult = int(_mult)
    spdf = pyfant.greek_to_spdf(_spdf)

    return label, mult, spdf


def _parse_A(browser, tag):
    """
    Parses the "A" information (A is the "coupling constant")

    The cell in column "Te" sometimes cites a reference; the anchor tag has an Id like "Dia19";
    the value we are looking for is in (roughly in the form "A = some_number"

    """
    for item in tag.children:
        if item.name == "a":
            try:
                id_ = re.match("#(Dia\d+)", item.get("href")).groups()[0]
                a = browser.find("a", id=id_)
                if a is not None:
                    td = a.next_element.next_element
                    stemp = re.match("<td>\s*A.*?=\s*([\(\)\+\-]*[0-9]+\.[0-9]+)", str(td)).groups()[0]
                    if stemp is None:
                        return None
                    try:
                        return float(stemp.replace("(", "").replace(")", ""))
                    except AttributeError:
                        # no match
                        return None
                    except ValueError:
                        # match, but no valid float
                        return None

            except AttributeError:
                pass


    return None

_conv_unicode_sup = {
    "i": "\u2071",
    "1": "\u00b9",
    "2": "\u00b2",
    "3": "\u00b3",
    "4": "\u2074",
    "+": "\u207a",
    "-": "\u207b"}

_conv_ascii_sup = {"1": "1",
                   "2": "2",
                   "3": "3",
                   "4": "4",
                   "+": "+",
                   "-": "-",}


_conv_unicode_img = {"larrow": "←",
                     "rarrow": "→",
                     "lrarrow": "↔"}

_conv_ascii_img = {"larrow": "<-",
                   "rarrow": "->",
                   "lrarrow": "<->"}


def _html_formula_to_plain(tag, flag_unicode):
    """
    Returns a string from a bs4.element.Tag representing a formula in NIST web book pages.

    Some HTML elements are converted to unicode characters
    """

    conv_sup = _conv_unicode_sup if flag_unicode else _conv_ascii_sup
    conv_img = _conv_unicode_img if flag_unicode else _conv_ascii_img

    parts = []
    for item in tag.children:
        if isinstance(item, bs4.element.NavigableString):
            parts.append(str(item))
        elif item.name == "sub":
            parts.append("_" + item.text)
        elif item.name == "sup":
            parts.append(conv_sup.get(item.text, "^" + item.text))
        elif item.name == "img":
            # parts.append("\\"+item.attrs["alt"])
            alt = item.attrs["alt"]
            parts.append(conv_img.get(alt, alt))
    return ("".join(parts)).strip().replace("  ", " ").replace("  ", " ")


def _floatify(str_):
    """Tries to convert to float, stripping brackets if it is the case"""
    try:
        ret = float(str_)
    except:
        # Note: It seems that NIST Chemistry WebBook has some OCR behind.
        #       I am making some effort to go around some cases I bumped into, for example, : "4.7E-_7"

        gg = re.match("[\[(]*([0-9\.\-E]+)", str_.replace("_", ""))
        if gg is None:
            ret = str_
        else:
            ret = float(gg.groups()[0])
    return ret


def _nonify(x):
    """Returns None if len(x) == 0, else x"""
    if isinstance(x, str) and len(x.strip()) == 0:
        return None
    return x






# todo plot "X" instead of line
# find line in editor by clicking
# rename molecule

__all__ = ["XFileAtoms"]

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT # as NavigationToolbar2QT
import matplotlib.pyplot as plt
import numpy as np
from .a_XAtomLinesEditor import *
import os.path
import webbrowser
import sys
from ._shared import *
import a99
import pyfant


NUM_PLOTS = len(ATOM_HEADERS)-1  # -1 because whe "lambda" does not have its plot

class XFileAtoms(QMainWindow):

    def __init__(self, parent=None):
        QMainWindow.__init__(self, parent)

        self.f = None  # FileAtoms object
        self.flag_sort = False
        self.atom_index = None
        self.atom = None  # Atom instance
        self.flag_changed = False
        self.save_dir = "."
        # Form with the table to edit the lines
        self.form_lines = None
        self.form_histogram = None

        # Information about the plots
        self.marker_row = None  # points into current atom, self.atom
        self.plot_info = [PlotInfo() for _ in range(NUM_PLOTS)]
        # keptself.set_flag_plot(ATOM_ATTR_NAMES.index("kiex")-1, True)
        self.set_flag_plot(ATOM_ATTR_NAMES.index("algf")-1, True)

        # ** "Atoms browser"

        # ** ** left of splitter
        self.labelAtoms = QLabel('Atoms list (Ctrl+1)')
        a = self.listWidgetAtoms = QListWidget()
        a.currentRowChanged.connect(self.on_listWidgetAtoms_currentRowChanged)
        # a.setEditTriggers(QAbstractItemView.DoubleClicked)
        # a.setEditTriggers(QAbstractItemView.AllEditTriggers)
        a.setContextMenuPolicy(Qt.CustomContextMenu)

        self.labelAtoms.setBuddy(self.listWidgetAtoms)

        l = self.layoutAtoms = QVBoxLayout()
        a99.set_margin(l, 0)
        l.setSpacing(1)
        l.addWidget(self.labelAtoms)
        l.addWidget(self.listWidgetAtoms)

        a = self.widgetAtoms = QWidget()
        a.setLayout(self.layoutAtoms)


        # ** ** right of splitter

        # ** ** ** ** Plot widget

        # ** ** ** ** ** Toolbar above plot

        am = self.buttonSort = QPushButton("Sort wave (Alt+&W)")
        am.clicked.connect(self.on_buttonSort_clicked)
        am.setCheckable(True)
        am.setToolTip("Sort spectral lines in ascending order of wavelength")
        if self.flag_sort:
            am.setChecked(True)

        # adds checkable buttons sj, jj
        self.plot_buttons = bb = []
        for i in range(NUM_PLOTS):  # sj, jj etc
            s = ATOM_HEADERS[i+1]
            b = QPushButton("%s (Alt+&%d)" % (s, i+1))
            b.clicked.connect(self.on_button_plot_clicked)
            b.setCheckable(True)
            if self.flag_plot(i):
                b.setChecked(True)
            bb.append(b)

        a11 = self.buttonEditLines = QPushButton("Edit lines (Ctrl+3)")
        a11.clicked.connect(self.on_buttonEditLines_clicked)
        a2 = self.labelNumLines = QLabel("--")
        a3 = self.spacer0 = QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum)

        l0 = self.layoutPlotToolbar = QHBoxLayout()
        l0.addWidget(am)
        for b in bb:
            l0.addWidget(b)
        l0.addWidget(a11)
        l0.addWidget(a2)
        l0.addItem(a3)
        a99.set_margin(l0, 1)
        a = self.widgetPlotToolbar = QWidget()
        a.setLayout(l0)
        a.setFixedHeight(40)

        # ** ** ** ** ** ** Plot widget

        # http://stackoverflow.com/questions/12459811
        self.figure = plt.figure()
        self.canvas = FigureCanvas(self.figure)
        self.canvas.mpl_connect('button_press_event', self.on_plot_click)
        self.toolbar = NavigationToolbar2QT(self.canvas, self)
        layout = QVBoxLayout()
        layout.addWidget(self.toolbar)
        layout.addWidget(self.canvas)
        a99.set_margin(layout, 0)

        a = self.widgetPlot = QWidget()
        a.setLayout(layout)

        l1 = self.layoutPlot = QVBoxLayout()
        l1.addWidget(self.widgetPlotToolbar)
        l1.addWidget(self.widgetPlot)
        a99.set_margin(l1, 0)

        a = self.widgetPlot = QWidget()
        a.setLayout(l1)


        # ** splitter: (list of Atoms) | (plot)
        a = self.splitter = QSplitter(Qt.Horizontal)
        a.addWidget(self.widgetAtoms)
        a.addWidget(self.widgetPlot)
        a.setStretchFactor(0, 2)
        a.setStretchFactor(1, 10)


        # * # * # * # * # * # * # *
        # Now the menu bar

        # self.menubar = QMenuBar(self)
        # self.menubar.setGeometry(QRect(0, 0, 772, 18))
        # self.menubar.setObjectName(_fromUtf8("menubar"))
        b = self.menuBar()
        m = self.menu_file = b.addMenu("&File")
        self.act_save = ac = m.addAction("&Save")
        ac.setShortcut("Ctrl+S")
        ac.triggered.connect(self.on_save)
        self.act_save_as = ac = m.addAction("Save &as...")
        ac.setShortcut("Ctrl+Shift+S")
        ac.triggered.connect(self.on_save_as)
        m.addSeparator()
        ac = m.addAction("&Quit")
        ac.setShortcut("Ctrl+Q")
        ac.triggered.connect(self.close)


        m = self.menu_tools = b.addMenu("&Tools")
        self.act_save = ac = m.addAction("&Histogram")
        ac.triggered.connect(self.on_histogram)

        # * # * # * # * # * # * # *
        # Final adjustments

        self.splitter.setFont(a99.MONO_FONT)
        self.setCentralWidget(self.splitter)
        a99.place_left_top(self)

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Qt override

    def closeEvent(self, event):
        if self.flag_changed:
            # http://straightedgelinux.com/blog/python/html/pyqtxt.html
            r = QMessageBox.question(self,
                        "About to exit",
                        "File \"%s\" has unsaved changes. Save now?" % self.f.filename,
                        QMessageBox.Yes | QMessageBox.No|
                        QMessageBox.Cancel)
            if r == QMessageBox.Cancel:
                event.ignore()
            elif r == QMessageBox.No:
                pass
            elif r == QMessageBox.Yes:
                try:
                    self.save()
                except:
                    # In case of error saving file, will not exit the program
                    event.ignore()
                    raise()
        if event.isAccepted():
            self.close_editor()

    def keyPressEvent(self, ev):
        if ev.modifiers() == Qt.ControlModifier:
            k = ev.key()
            if k == Qt.Key_1:
                self.listWidgetAtoms.setFocus()
            elif k == Qt.Key_3:
                self.on_buttonEditLines_clicked()

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Slots

    def on_help(self, _):
        base_dir = os.path.dirname(sys.argv[0])
        try:
            webbrowser.open_new(os.path.join(base_dir, "ated.html"))
            a99.show_message("Help file ated.html was opened in web browser.")
        except Exception as e:
            a99.show_error(str(e))
            raise

    def on_save(self, _):
        self.disable_save_actions()
        try:
            self.save()
        except Exception as e:
            a99.show_error(str(e))
            raise
        finally:
            self.enable_save_actions()

    def on_save_as(self, _):
        self.disable_save_actions()
        try:
            if self.f:
                new_filename = QFileDialog.getSaveFileName(self, "Save file",
                 self.save_dir, "*.dat")[0]
                if new_filename:
                    self.save_dir, _ = os.path.split(str(new_filename))
                    self.save_as(new_filename)
        except Exception as e:
            a99.show_error(str(e))
            raise
        finally:
            self.enable_save_actions()

    def on_listWidgetAtoms_currentRowChanged(self, row):
        if row > -1:
            self.set_atom(row)

    def on_button_plot_clicked(self):
        # This may be triggered by any of the plot buttons
        button = self.sender()
        idx = self.plot_buttons.index(button)
        self.set_flag_plot(idx, button.isChecked())
        self.plot_lines()

    def on_buttonSort_clicked(self):
        self.flag_sort = self.buttonSort.isChecked()
        self.plot_lines()

    def on_histogram(self, _):
        if self.form_histogram is None:
            self.form_histogram = pyfant.XFileAtomsHistogram(self.f)
        self.form_histogram.show()


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #

    def load(self, f):
        """Loads file into GUI."""
        assert isinstance(f, pyfant.FileAtoms)

        self.f = f

        for m in f.atoms:
            assert isinstance(m, pyfant.Atom)
            item = QListWidgetItem(self.get_atom_string(m))
            # not going to allow editing yet item.setFlags(item.flags() | Qt.ItemIsEditable)
            self.listWidgetAtoms.addItem(item)

        if len(f.atoms) > 0:
            self.listWidgetAtoms.setCurrentRow(0)

        self.flag_changed = False
        self.update_window_title()

    def save(self):
        if self.f:
            self.f.save_as()
            self.flag_changed = False
            self.update_window_title()

    def save_as(self, filename):
        if self.f:
            self.f.save_as(filename)
            self.flag_changed = False
            self.update_window_title()

    def set_atom(self, i):
        self.atom_index = i
        m = self.atom = self.f.atoms[i]
        self.update_atom_info()
        self.marker_row = None
        self.plot_lines()
        self.set_editor_atom()

    def plot_lines(self):
        self.clear_markers()
        o = self.atom
        if o is not None:
            self.figure.clear()

            n = sum([info.flag for info in self.plot_info])  # number of subplots (0, 1 or 2)
            # map to reuse plotting routine, contains what differs between each plot
            map_ = [(ATOM_HEADERS[i], o.__getattribute__(ATOM_ATTR_NAMES[i]))
                    for i in range(1, len(ATOM_HEADERS))]

            # number of rows and columns for each different number of subplots
            SL = [(1, 1), (2, 1), (2, 2), (2, 2), (3, 2), (3, 2)]

            i_subplot = 1
            for i in range(len(map_)):
                y_label = map_[i][0]
                pi = self.plot_info[i]
                pi.y_vector = _y = map_[i][1]

                if pi.flag:
                    if not self.flag_sort:
                        x = o.lambda_
                        y = _y
                    else:
                        _x = np.array(o.lambda_)
                        _y = np.array(_y)
                        ii = np.argsort(_x)
                        x = _x[ii]
                        y = _y[ii]

                    a99.format_BLB()

                    self.figure.add_subplot(SL[n-1][0], SL[n-1][1], i_subplot)
                    pi.axis = ax = self.figure.gca()
                    ax.clear()
                    ax.plot(x, y, 'k'+('' if len(x) > 1 else 'x'))
                    ax.set_xlabel('Wavelength ($\AA$)')
                    ax.set_ylabel(y_label)

                    # x-limits
                    xmin, xmax = min(x), max(x)
                    K = .02*(xmax-xmin)
                    ax.set_xlim([xmin-K, xmax+K])

                    # y-limits
                    ymin, ymax = min(y), max(y)
                    K = .02*(ymax-ymin)
                    ax.set_ylim([ymin-K, ymax+K])

                    i_subplot += 1

            if i_subplot > 1:
                plt.tight_layout()

            self.canvas.draw()
            self.draw_markers()

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #

    def update_atom_info(self):
        """Updates the "number of lines" label."""
        a = self.atom
        n = len(a)
        self.labelNumLines.setText('Number of lines: %d' % (n,))

    def update_window_title(self):
        self.setWindowTitle("ated -- %s" %
                            (self.f.filename+("" if not self.flag_changed else " (changed)"), ))

    def enable_save_actions(self):
        self.act_save.setEnabled(True)
        self.act_save_as.setEnabled(True)

    def disable_save_actions(self):
        self.act_save.setEnabled(False)
        self.act_save_as.setEnabled(False)


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Routines related to the lines editor

    def on_buttonEditLines_clicked(self):
        if self.form_lines is None:
            f = self.form_lines = XAtomLinesEditor(self)
            f.show()
            self.set_editor_atom()
        self.plot_lines()

    def AtomLinesEditor_closing(self):
        """Called by the atomic lines editor to notify that it is closing."""
        self.form_lines = None
        self.marker_row = None
        self.plot_lines()  # to remove the "X" from the plot

    def AtomLinesEditor_current_row_changed(self, currentRow):
        """Called by the atomic lines editor to notify that the current row has changed."""
        self.set_marker_row(currentRow)

    def AtomLinesEditor_cell_changed(self, row, column, value):
        """Called by the atomic lines editor to notify that a value has changed."""
        attr_name = ATOM_ATTR_NAMES[column]
        old_value = self.atom.lines[row].__getattribute__(attr_name)
        if old_value != value:
            self.atom.lines[row].__setattr__(attr_name, value)
            self.flag_changed = True
            self.plot_lines()
            self.update_window_title()

    def set_editor_atom(self):
        if self.atom is not None and self.form_lines is not None:
            self.form_lines.set_atom(self.atom, "Atom: "+self.listWidgetAtoms.currentItem().text())

    def close_editor(self):
        if self.form_lines is not None:
            self.form_lines.close()

    # Controlling plots and markers for current row

    def set_marker_row(self, i):
        self.clear_markers()
        self.marker_row = i
        self.draw_markers()

    def clear_markers(self):
        for o in self.plot_info:
            if o.mpl_obj:
                a99.remove_line(o.mpl_obj)
                o.mpl_obj = None

    def draw_markers(self):
        self.clear_markers()
        if self.marker_row is not None and any([o.flag for o in self.plot_info]):
            i = self.marker_row
            lambda_ = self.atom.lambda_[i]
            for o in self.plot_info:
                if o.flag:
                    # http://stackoverflow.com/questions/22172565/matplotlib-make-plus-sign-thicker
                    o.mpl_obj = o.axis.plot([lambda_], [o.y_vector[i]], 'xr', mew=2, ms=10)
            self.canvas.draw()

    def flag_plot(self, idx):
        return self.plot_info[idx].flag

    def set_flag_plot(self, idx, x):
        self.plot_info[idx].flag = x

    def on_plot_click(self, event):
        lambda_ = event.xdata
        if lambda_ is not None and self.form_lines is not None:
            idx = a99.index_nearest(self.atom.lambda_, lambda_)
            self.form_lines.set_row(idx)
            # self.set_marker_row(idx)

            # print 'button=%d, x=%d, y=%d, xdata=%f, ydata=%f'%(
            #     event.button, event.x, event.y, event.xdata, event.ydata)

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #

    @staticmethod
    def get_atom_string(a):
        assert isinstance(a, pyfant.Atom)
        return "%-3s (%4d)" % (str(a).strip(), len(a))



"""Plez atomic or molecular lines file"""

__all__ = ["FilePlezTiO", "PlezTiOLine"]

# from ..gear import *
import sys
import a99
from f311 import DataFile
from collections import namedtuple
import struct
import os
import numpy as np

#: List of all atomic symbols
_symbols = [
'H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na', 'Mg', 'Al', 'Si',
 'P', 'S', 'Cl', 'Ar', 'K', 'Ca', 'Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co',
 'Ni', 'Cu', 'Zn', 'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr', 'Rb', 'Sr', 'Y', 'Zr',
 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd', 'In', 'Sn', 'Sb', 'Te', 'I',
 'Xe', 'Cs', 'Ba', 'La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy',
 'Ho', 'Er', 'Tm', 'Yb', 'Lu', 'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt', 'Au',
 'Hg', 'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac', 'Th', 'Pa', 'U',
 'Np', 'Pu', 'Am', 'Cm', 'Bk', 'Cf', 'Es', 'Fm', 'Md', 'No', 'Lr', 'Rf', 'Db',
 'Sg', 'Bh', 'Hs', 'Mt'
]

# Position
#           1         2         3         4         5         6         7         8         9        10        11
# 01234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678
#
#    lambda_air(A)   gf         Elow(cm-1)  vl  Jl    Nl  syml  Eup(cm-1) vu   Ju    Nu  symu  gamrad    mol trans branch
#      3164.8716 0.769183E-11   3459.6228   0   5.0   5.0  0  35047.3396  15   6.0   6.0  0 1.821557E+07 'TiO a f  R    '
#      3164.8764 0.931064E-11   3466.0556   0   6.0   6.0  0  35053.7238  15   7.0   7.0  0 1.818709E+07 'TiO a f  R    '
#      3164.8827 0.609470E-11   3454.2620   0   4.0   4.0  0  35041.8672  15   5.0   5.0  0 1.824816E+07 'TiO a f  R    '




PlezTiOLine = namedtuple("PlezTiOLine", ["lambda_", "gf", "Elow", "vlow", "Jlow",
    "Nlow", "symlow", "Eup", "vup", "Jup", "Nup", "symup", "gamrad", "mol", "state_from", "state_to", "branch"])
PlezTiOLine.__doc__ = """Represents one Plez molecular line

Field names were maintained as in Plez 'ReadmeTiO' file with slight changes.
"""


class FilePlezTiO(DataFile):
    """
    Plez molecular lines file, TiO format

    Lines are encoded as PlezTiOLine object

    Because Plez files are too big, they will be ***filtered upon load** (unfiltered files may take up
    several GB of RAM memory if fully loaded). See arguments list

    Args:
        min_gf=0: minimum gf
        max_J2l=99999: Maximum J" (J lower)
        max_vl=99999: Maximum v' (v upper)
        max_v2l=99999: Maximum v" (v lower)

    Example:

        >>> import pyfant
        >>> f = pyfant.FilePlezTiO(min_gf=10**-4, max_J2l=120, max_vl=9, max_v2l=9)
        >>> f.load("linelist_reduced46_all_deltacorr.dat")
        >>> print("Number of lines loaded: {}".format(len(f))
    """

    attrs = ["num_lines"]

    @property
    def num_lines(self):
        return len(self)

    def __len__(self):
        return len(self.lines)

    def __init__(self, min_gf=0, max_J2l=99999, max_vl=99999, max_v2l=99999):
        DataFile.__init__(self)
        self.lines = []
        self.min_gf = min_gf
        self.max_J2l = max_J2l
        self.max_vl = max_vl
        self.max_v2l = max_v2l

    def __iter__(self):
        return iter(self.lines)



    def get_numpy_array(self):
        """Returns a np array with named fields, same names as in PlezTiOLine"""
        dtype = self._get_dtype()
        return np.array(self.lines, dtype=dtype)

    def _get_dtype(self):
        # dtype = [("lambda_", float),
        #          ("gf", float),
        #          ("Elow", float),
        #          ("vlow", int),
        #          ("Jlow", float),
        #          ("Nlow", float),
        #          ("symlow", int),
        #          ("Eup", float),
        #          ("vup", int),
        #          ("Jup", float),
        #          ("Nup", float),
        #          ("symup", int),
        #          ("gamrad", float),
        #          ("mol", '|S3'),
        #          ("state_from", '|S1'),
        #          ("state_to", '|S1'),
        #          ("branch", '|S6'), ]
        dtype = [("lambda_", float),
                 ("gf", float),
                 ("Elow", float),
                 ("vlow", int),
                 ("Jlow", float),
                 ("Nlow", float),
                 ("symlow", int),
                 ("Eup", float),
                 ("vup", int),
                 ("Jup", float),
                 ("Nup", float),
                 ("symup", int),
                 ("gamrad", float),
                 ("mol", '|S3'),
                 ("state_from", '|S1'),
                 ("state_to", '|S1'),
                 ("branch", '|S6'), ]
        return dtype

    def _do_load(self, filename):
        def strip(s): return s.decode("ascii").strip()
        def I(s): return s.decode("ascii")  # Identity
        my_struct = struct.Struct("14s 13s 12s 4s 6s 6s 3s 12s 4s 6s 6s 3s 13s 2x 3s 1x 1s 1x 1s 6s 2x")
        func_map = [float, float, float, int, float, float, int, float, int, float, float, int,
                    float, strip, I, I, strip]

        filesize = os.path.getsize(filename)
        num_lines = float(filesize)/120

        if num_lines != int(num_lines):
            raise RuntimeError("Fractionary number of lines: {}, not a FilePlezTiO".format(num_lines))
        num_lines = int(num_lines)

        with open(filename, "rb") as h:
            try:
                r = 0  # counts rows of file
                ii = 0  # progress feedback auxiliary variable
                while True:
                    s = h.readline()

                    # print(s)
                    # break
                    if len(s) == 0: break  # # EOF: blank line or references section

                    # Filtering part

                    gf = float(s[14:27])
                    J2l = float(s[43:49])
                    vl = int(s[70:74])
                    v2l = int(s[49:53])

                    # gf = float(bits[1])
                    # J2l = float(bits[4])
                    # vl = float(bits[8])
                    # v2l = float(bits[3])

                    # gf = args[1]
                    # J2l = args[4]
                    # vl = args[8]
                    # v2l = args[3]


                    if not (gf >= self.min_gf and J2l <= self.max_J2l and vl <= self.max_vl and v2l <= self.max_v2l):
                        r +=1
                        continue

                    bits = my_struct.unpack_from(s)

                    args = [func(x) for func, x in zip(func_map, bits)]
                    line = PlezTiOLine(*args)
                    self.lines.append(line)
                    r += 1

                    # args = [func(x) for func, x in zip(func_map, bits)]
                    # line = PlezTiOLine(*bits)
                    # self.lines.append(line)



                    #           1         2         3         4         5         6         7         8         9        10        11
                    # 01234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678
                    #
                    #    lambda_air(A)   gf         Elow(cm-1)  vl  Jl    Nl  syml  Eup(cm-1) vu   Ju    Nu  symu  gamrad    mol trans branch
                    #      3164.8716 0.769183E-11   3459.6228   0   5.0   5.0  0  35047.3396  15   6.0   6.0  0 1.821557E+07|'TiO a f  R    '
                    # 12345678901234                                                                                        | 12345678901234
                    # 0             1234567890123                                                                           |
                    #               1            123456789012                                                               |
                    #                            2           1234                                                           |
                    #                                        3   123456                                                     |
                    #                                            4     123456                                               |
                    #                                                  5     123                                            |
                    #                                                        6  123456789012                                |
                    #                                                           7           1234                            |
                    #                                                                       8   123456                      |
                    #                                                                           9     123456                |
                    #                                                                                 10    123             |
                    #                                                                                       11 1234567890123|
                    #                                                                                          12           2x
                    #                                                                                                         123 1 1 123456
                    #                                                                                                         13  141516

                    ii += 1
                    if ii > 12345:
                        a99.get_python_logger().info(
                            "Loading '{}': {}".format(filename, a99.format_progress(r + 1, num_lines)))
                        ii = 0


            except Exception as e:
                raise type(e)(("Error around %d%s row of file '%s'" %
                               (r + 1, a99.ordinal_suffix(r + 1), filename)) + ": " + str(
                    e)).with_traceback(sys.exc_info()[2])



    def _do_save_as(self, filename):
        with open(filename, "w") as h:
            for line in self.lines:
               h.write("{:14.4f} {:12e} {:11.4f} {:3d} {:5.1f} {:5.1f} {:2d} {:11.4f} {:3d} {:5.1f} {:5.1f} {:2d} {:12e} '{:3s} {:1s} {:1s} {:6s}'\n".format(*line))



#!/usr/bin/env python
"""
Plots interpolated atmospheric model curve (hopefully) in the middle of the curves
used by innewmarcs for interpolation.

The file containing which files and records innewmarcs chose is assumed to be called
'innewmarcs_explain.txt'

This file can be generated by invoking the Fortran binary like this:
  > innewmarcs --explain
"""


import argparse
import matplotlib.pyplot as plt
import re
import numpy as np
from itertools import cycle
from mpl_toolkits.mplot3d import Axes3D  # yes, required
import a99
import logging
import pyfant


a99.logging_level = logging.INFO
a99.flag_log_file = True


def parse_explain_file(filename):
    """Extracts information from "explain" file created by innewmarcs."""
    tokens = {}
    with open(filename) as h:
        while True:
            s = h.readline().strip()
            if not s:
                break
            m = re.match(r"(.+):\s*(.*)", s)
            if not m:
                raise RuntimeError('Error parsing file "%s"' % filename)
            token, s_data = m.groups()
            tokens[token] = eval(s_data)
    return tokens


FN_EXPLAIN = 'innewmarcs_explain.txt'
VARS = ['nh', 'teta', 'pe', 'pg', 'log_tau_ross']


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
     description=__doc__,
     formatter_class=a99.SmartFormatter
     )
    parser.add_argument('--inum', type=int, default=1, help='Record number (>= 1)')
    parser.add_argument('--var', type=str, help='Variable to plot', default=VARS[0], choices=VARS)

    args = parser.parse_args()

    tokens = parse_explain_file(FN_EXPLAIN)
    fn_grid = tokens['innewmarcs grid file']
    fn_output = tokens['innewmarcs output file']
    indexes = np.array(np.matrix(tokens['innewmarcs indexes'])).flatten()
    records = []  # ModRecord, name
    m = pyfant.FileModBin()
    m.load(fn_grid)
    for i in indexes:
        record_name = '%s#%d' % (fn_grid, i)
        records.append((m.records[i-1], record_name))

    mod = pyfant.FileModBin()
    mod.load(fn_output)
    rm = mod.records[0]

    print_record = lambda record, name: print(name, '; teff=', record.teff, '; glog=', record.glog, '; asalog=', record.asalog)

    # Figure
    plt.figure()
    for record, name in records:
        print_record(record, name)
        plt.plot(record.log_tau_ross, record.__getattribute__(args.var), label=name)
    name = fn_output
    v = rm.__getattribute__(args.var)
    if any(np.isnan(v)):
        print("NaN ALERT!!!")
    print_record(rm, name)
    plt.plot(rm.log_tau_ross, v, label=name, linewidth=2, linestyle='--')
    plt.xlabel('log_tau_ross')
    plt.title(args.var)
    plt.legend(loc=0)

    # Scatter figure
    fig = plt.figure()
    ax = fig.gca(projection='3d')
    markers = cycle("ov^<>p*+x")
    for r, name in records:
        ax.scatter([r.teff], [r.glog], [r.asalog],
                   c=next(ax._get_lines.color_cycle), s=60,
                   label=name, marker=next(markers))
    ax.scatter([rm.teff], [rm.glog], [rm.asalog],
               c=next(ax._get_lines.color_cycle), s=60,
               label=fn_output, marker=next(markers))
    ax.set_xlabel('teff')
    ax.set_ylabel('glog')
    ax.set_zlabel('asalog')
    fig.canvas.set_window_title('teff-glog-asalog scatterplot')
    plt.legend(loc=0)

    # Another figure
    fig = plt.figure()
    y_var_names = ["teff", "glog", "asalog"]
    for i, y_var_name in enumerate(y_var_names):
        ax = fig.add_subplot(len(y_var_names), 1, i+1, projection='3d')
        x = None
        for record, name in records:
            print_record(record, name)
            v = record.__getattribute__(args.var)
            x = record.log_tau_ross
            y_var = record.__getattribute__(y_var_name)
            y = np.ones((len(x),))*y_var
            ax.plot(x, y, v, label=name)

        name = fn_output
        v = rm.__getattribute__(args.var)
        x = rm.log_tau_ross
        y_var = rm.__getattribute__(y_var_name)
        y = np.ones((len(x),))*y_var
        print_record(rm, name)
        plt.plot(x, y, v, label=name, linewidth=2, linestyle='--')

        plt.xlabel('log_tau_ross')
        plt.ylabel(y_var_name)
        plt.gca().set_zlabel(args.var)
        plt.title("Layer X %s X %s" % (y_var_name, args.var))
        plt.legend(loc=0)

    plt.show()




from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from a99 import WDBRegistry
import a99


__all__ = ["WDBRegistry"]

class WDBMolecule(WDBRegistry):
    """Registry for table 'molecule'"""


    def __init__(self, *args):
        WDBRegistry.__init__(self, *args)

    # # Override

    def _f_changed(self):
        self._populate()
#        self._move_to_first()

    def _find_formula(self, formula):
        """Moves to row where formula is (if found, otherwise does nothing)"""
        for i, row in enumerate(self._data):
            if row["formula"] == formula:
                self.tableWidget.setCurrentCell(i, 0)
                break

    def _populate(self, restore_mode=None):
        """
        Populates table widget

        Args:
            restore_mode: how to reposition cursor:
                "formula": tries to re-position at same formula as before
                "index": tries to restore same row index
                anything else: does not re-position
        """

        if self._f is None:
            return

        self._flag_populating = True
        try:
            t = self.tableWidget
            if self._f is None:
                a99.reset_table_widget(t, 0, 0)
                return

            curr_idx = self.tableWidget.currentRow()
            curr_row = self.row

            fieldnames = list(self._f.get_table_info("molecule"))
            rows = a99.cursor_to_rows(self._f.query_molecule())
            nr, nc = len(rows), len(fieldnames)
            a99.reset_table_widget(t, nr, nc)
            t.setHorizontalHeaderLabels(fieldnames)
            if nr > 0:
                for i, row in enumerate(rows):
                    for j, fieldname in enumerate(fieldnames):
                        cell = row[fieldname]
                        item = QTableWidgetItem("" if cell is None else str(cell))
                        t.setItem(i, j, item)
            t.resizeColumnsToContents()
            self._data = rows

            if restore_mode == "formula":
                if curr_row:
                    self._find_formula(curr_row["formula"])
            elif restore_mode == "index":
                if -1 < curr_idx < t.rowCount():
                    t.setCurrentCell(curr_idx, 0)
        finally:
            self._flag_populating = False
            self._wanna_emit_id_changed()


    def _get_edit_params(self):
        """Returns a Parameters object containing information about the fields that may be edited"""
        ti = self._f.get_table_info("molecule")
        params = a99.table_info_to_parameters(ti)
        params = [p for p in params if not p.name.startswith("id")]
        return params


    def _get_edit_field_names(self):
        """Returns a list with the names of the fields that may be edited"""
        params = self._get_edit_params()
        ret = [p.name for p in params]
        return ret


    def _do_on_insert(self):
        params = self._get_edit_params()
        form = a99.XParametersEditor(specs=params, title="Insert molecule")
        r = form.exec_()
        if r == QDialog.Accepted:
            kwargs = form.get_kwargs()
            conn = self._f.get_conn()
            s = "insert into molecule({}) values ({})".format(", ".join([p.name for p in params]),
                                                              ", ".join(["?"]*len(params)))
            cursor = conn.execute(s, list(kwargs.values()))
            id_ = cursor.lastrowid
            conn.commit()
            self._populate(False)
            self._find_id(id_)
            return True


    def _do_on_edit(self):
        r, form = a99.show_edit_form(self.row, self._get_edit_field_names(), "Edit molecule")
        if r == QDialog.Accepted:
            kwargs = form.get_kwargs()
            s = "update molecule set {} where id = {}".format(
                ", ".join(["{} = '{}'".format(a, b) for a, b in kwargs.items()]), self.row["id"])
            conn = self._f.get_conn()
            conn.execute(s)
            conn.commit()
            self._populate()
            self._find_formula(kwargs["formula"])
            return True


    def _do_on_delete(self):
        r = QMessageBox.question(None, "Delete molecule",
                                 "This will also delete all States, Systems, and PFANT molecules associated to this molecule!\n\nAre you sure?",
                                 QMessageBox.Yes | QMessageBox.No, QMessageBox.No)
        if r == QMessageBox.Yes:
            conn = self._f.get_conn()
            id_ = self.row["id"]

            for row in conn.execute("select id from system where id_molecule = ?", [id_]):
                conn.execute("delete from fcf where id_system = ?", [row["id"]])
            conn.execute("delete from molecule where id = ?", [id_])
            conn.execute("delete from state where id_molecule = ?", [id_])
            conn.execute("delete from system where id_molecule = ?", [id_])
            conn.execute("delete from pfantmol where id_molecule = ?", [id_])
            conn.commit()

            self._populate("index")
            return True



from .calc_qgbd import *
import datetime
from collections import OrderedDict
import math
import pyfant
import a99
from enum import Enum

__all__ = ["Conv", "ConvMode"]


# _DEFAULT_QGBD_CALCULATOR = calc_qgbd_tio_like


class ConvMode(Enum):
    # Hönl-London factors
    HLF = 1
    # Einstein's "A" coefficient (1/s)
    EINSTEIN = 2
    EINSTEIN_MINIMAL = 3
    # Oscillator strength (f_(Jl,J2l)-falue)
    F = 4


class Conv(object):
    """Molecular lines converter base class

    Args:
        mode: specifies how SJ's are calculated:
                ConvMode.HLF: using Honl-London factors using Kovácz formulation
                ConvMode.EINSTEIN: using Einstein's "A" coefficients (given in s^-1)
                ConvMode.EINSTEIN_MINIMAL: using Einstein's "A" coefficients (given in s^-1); and
                                           assuming vl=0, v2l=0, J2l=Jl. Experimental conversion using minimal data
                                           from Plez linelist

                **Note** mode is handles by subclasses and may be partially implemented or ignored. The code
                         (e.g. conv_brooke2014.py) must be checked for class-specific behaviour

        molconsts: a dict-like object combining field values from tables 'molecule', 'state',
                    'pfantmol', and 'system' from a FileMolDB database

        flag_norm_sj: Whether to multiply calculated SJ's by normalization factor

        fcfs: Franck-Condon Factors (dictionary of floats indexed by (vl, v2l))

        flag_quiet: Will not log exceptions when a molecular line fails

        flag_special_fcf: check _get_fcf() for formula: transforms any number <n> into 0.<n>; no idea why. This was some
                          Kurucz conversion 2017 thing

        molcomment: comments to be added to molecule description

        fe: (=None) if used, replaces "fe" from molecular constants

        do: (=None) if used, replaces "do" from molecular constants

        fact: (=None) if used, replaces default pyfant.SetOfLines.fact

        strengthfactor: (=1.) all resulting line strengths ("sj") will be multiplied by this factor
                        Note: it is better to use "fe" argument instead
        flag_fcf: Whether to multiply calculated gf's by Franck-Condon Factor
                  (only makes sense when mode==ConvMode.HLF)

        flag_filter_labels: whether to filter lines by labels [taken from self.molconsts e.g. ("A", "X")].

        moldb: FileMolDB instance to automatically get information from (for example will take FCF's if needed)

        inputfilename: if passed, will be included in FileMolecules.titm

    Both strengthfactor and fe are ways to achieve the same result (multiply the line strength), but the former applies
    to sj whereas the latter is recorded as the molecule's "FE"
    """

    def __init__(self, mode=ConvMode.HLF, molconsts=None, flag_norm_sj=True, fcfs=None,
                 flag_quiet=True, fe=None, do=None, fact=None,
                 flag_special_fcf=False, molcomment="", strengthfactor=1.,
                 flag_fcf=False, flag_filter_labels=False, moldb=None, inputfilename=None):
        self.mode = mode
        # self.qgbd_calculator = qgbd_calculator if qgbd_calculator else calc_qgbd_tio_like
        self.molconsts = molconsts
        self.fcfs = fcfs
        self.flag_norm_sj = flag_norm_sj
        self.flag_quiet = flag_quiet
        self.molcomment = molcomment
        self.fe = fe
        self.do = do
        self.fact = fact
        self.flag_special_fcf = flag_special_fcf
        self.strengthfactor = strengthfactor
        self.flag_fcf = flag_fcf
        self.flag_filter_labels = flag_filter_labels
        self.inputfilename = inputfilename

        if flag_fcf and fcfs is None:
            if not moldb:
                raise ValueError("flag_fcf==True, fcfs==None but moldb not provided to automatically get Franck-Condon factors!")
            self.fcfs = moldb.get_fcf_dict(molconsts["id_system"])
            if len(self.fcfs) == 0:
                raise RuntimeError("Frank-Condon factors not available")

        # These will be filled in upon conversion
        self.sols = None
        self.log = None
        self.mtools = None

    def make_file_molecules(self, lines):
        """
        Builds a FileMolecules object

        Args:
            lines: in most cases, a DataFile instance, but may be anything, since the molecular
                   lines data extraction is performed by a particular descendant class.
                   For example, the HITRAN converter expects a structure from the HAPI

        Returns:
            f, log: FileMolecules, MolConversionLog instances
        """

        # Runs specific conversor to SetOfLines
        self.sols = {}
        self.log = pyfant.MolConversionLog()
        self._make_sols(lines)

        sols_list = list(self.sols.values())
        sols_list.sort(key= lambda sol: sol.vl*1000+sol.v2l)

        mol = pyfant.molconsts_to_molecule(self.molconsts)

        if self.fe:
            mol.fe = self.fe
        if self.do:
            mol.do = self.do
        if self.molcomment:
            mol.description += f"; {self.molcomment}"

        mol.sol = sols_list
        f = pyfant.FileMolecules()
        now = a99.dt2str(datetime.datetime.now())

        fromfile = ""
        if self.inputfilename:
            fromfile = f" from file '{self.inputfilename}'"

        f.titm = f"Converted using pyfant.{self.__class__.__name__} @ {now}{fromfile}"
        f.molecules = [mol]
        return f

    def kovacs_toolbox(self):
        """Wraps f311.physics.multiplicity.kovacs_toolbox()"""
        return pyfant.kovacs_toolbox(self.molconsts, flag_normalize=self.flag_norm_sj)

    def append_line(self, line, gf_pfant, branch):
        """Use to append line to object"""
        self.append_line2(line.vl, line.v2l, line.lambda_, gf_pfant, line.J2l, branch)

    def append_line2(self, vl, v2l, lambda_, sj, jj, branch):
        """Alernative way to append line, created in 2023. Takes (vl, v2l) and PFANT relevant line-specific info.

        append_line() requires a "line" object which made sense when ConvKurucz was implemented, but unelegant when one
        already has all the values loose as variables.

        Arguments are as in PFANT molecular linelist files. See filemolecules.py and pfantlib.f90::file_molecules
        module.
        """
        sol_key = (vl, v2l)

        if self.flag_fcf and sol_key not in self.fcfs:
            raise KeyError("FCF not available for (vl, v2l) = ({}, {})".format(vl, v2l))

        if sol_key not in self.sols:
            qv = self.fcfs[sol_key] if self.flag_fcf else 1.
            gbd = calc_gbd(self.molconsts, v2l)  # self.qgbd_calculator(self.molconsts, v2l)
            self.sols[sol_key] = pyfant.SetOfLines(vl, v2l, qv, gbd["gv"], gbd["bv"], gbd["dv"],
                                                   fact=1. if self.fact is None else self.fact)

        self.sols[sol_key].append_line(lambda_, sj*self.strengthfactor, jj, branch)

        self.log.cnt_in += 1

    # Must reimplement this
    def _make_sols(self, lines):
        """Converts molecular lines into a list of SetOfLines object

        Args:
            lines: see make_file_molecules()

        Returns:
            None
        """
        raise NotImplementedError()

    ######
    # Gear

    def get_hlf(self, vl, v2l, J2l, branch):
        """Calculates SJ Hönl-London factor according to Kovács formulation.

        Returns:
            sj
        """

        if self.mtools is None:
            self.mtools = self.kovacs_toolbox()

        sj = self.mtools.get_sj(vl, v2l, J2l, branch)
        if sj < 0:
            raise ValueError("Negative SJ")

        return sj

    @staticmethod
    def get_sj_einstein(A, Jl, J2l, S2l, deltak, nu):
        """Calculates SJ and wavelength using Einstein's coefficient "A"."""
        normalizationfactor = Conv.get_normalizationfactor(J2l, S2l, deltak)
        SJ = normalizationfactor*A*1.499*(2*Jl+1)/nu**2
        return SJ

    @staticmethod
    def get_normalizationfactor(J2l, S2l, deltak):
        ret = 1/((2*S2l+1)*(2*J2l+1)*(2-deltak))
        return ret

    # (20230719) This is not used ATM
    def _get_fcf(self, vl, v2l):
        try:
            fcf = self.fcfs[(vl, v2l)]

            if self.flag_special_fcf:
                fcf = fcf * 10 ** -(math.floor(math.log10(fcf)) + 1)

        except KeyError as e:
            raise KeyError("FCF not available for (vl, v2l) = ({}, {})".format(vl, v2l))
        return fcf



########################################################################################################################

def calc_gbd(molconsts, v_lo):
    """
    Calculates gv, bv, dv

    Based on Fortran source 'agrup.plez7.f'

    Args:
        molconsts: dict-like object with keys:
            "omega_e"
            "omega_ex_e"
            "omega_ey_e"
            "B_e"
            "alpha_e"
            "D_e"
            "beta_e"

        v_lo: (integer) low transition level

    Returns: bdg, i.e., a dictionary with keys:
            "gv", "bv", "dv", "gzero"
    """

    omega_e = molconsts["state2l_omega_e"]
    omega_ex_e = molconsts["state2l_omega_ex_e"]
    omega_ey_e = molconsts["state2l_omega_ey_e"]
    B_e = molconsts["state2l_B_e"]
    D_e = molconsts["state2l_D_e"]
    alpha_e = molconsts["state2l_alpha_e"]
    beta_e = molconsts["state2l_beta_e"]

    if int(v_lo) != v_lo:
        raise ValueError("Argument 'v_lo' must be an integer")

    gzero = omega_e / 2.0 - omega_ex_e / 4.0 + omega_ey_e / 8.0
    v_lo5 = v_lo + 0.5

    # rotational constant "B_v"
    bv = B_e - alpha_e * v_lo5
    # rotational constant "D_v"
    dv = (D_e + beta_e * v_lo5) * 1.0e+06
    # rotational term " G_A"
    gv = omega_e * v_lo5 - omega_ex_e * v_lo5 ** 2 + omega_ey_e * v_lo5 ** 3 - gzero

    gbd = {"gv": gv, "bv": bv, "dv": dv, "gzero": gzero}

    return gbd



from . import basic
import os
import inspect
import a99
import pyfant
import re

__all__ = ["MolConsts", "some_molconsts", "MolConstPopulateError",
]

class MolConsts(dict):
    """Dict subclass that will hold several molecular constants, e.g., a table join"""

    # dictionary keys already present (fallback to None) at __init__()
    _KEYS = ["fe", "bm", "te", "do", "ua", "am", "ub",
             "from_label", "from_mult", "from_spdf", "to_label", "to_mult", "to_spdf",
             "statel_omega_e", "statel_B_e", "statel_beta_e", "statel_omega_ex_e", "statel_alpha_e",
             "statel_A", "statel_omega_ey_e", "statel_D_e",
             "state2l_omega_e", "state2l_B_e", "state2l_beta_e", "state2l_omega_ex_e",
             "state2l_alpha_e",
             "state2l_A", "state2l_omega_ey_e", "state2l_D_e", "name", "formula",
             "id_molecule", "id_pfantmol", "id_system", "id_statel", "id_state2l",
             "pfant_name", "pfant_notes"]

    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)

        for key in self._KEYS:
            self.setdefault(key, None)

    def __repr__(self):
        """Will generate sth like 'MolConsts({...})'"""
        _code = "{}({})".format(self.__class__.__name__, dict.__repr__(self))
        code = a99.make_code_readable(_code)
        return code

    def None_to_zero(self):
        """Replaces None values with zero"""

        for key in self:
            if self[key] is None:
                self[key] = 0.

    def get_str(self, style=basic.SS_PLAIN):
        """Returns string such as 'OH [A 2 SIGMA - X 2 PI]

        Args:
            style: SS_*
        """
        need = ("formula", "from_label", "from_mult", "from_spdf", "to_label", "to_mult", "to_spdf")
        self._i_need(need)

        return "{} {}".format(self["formula"], basic.molconsts_to_system_str(self, style))

    def get_system_str(self, style=basic.SS_PLAIN):
        """Returns string such as 'A 2 SIGMA - X 2 PI

        Args:
            style: SS_*
        """
        need = ("from_label", "from_mult", "from_spdf", "to_label", "to_mult", "to_spdf")
        self._i_need(need)

        return basic.molconsts_to_system_str(self, style)

    def get_S2l(self):
        """**Convention** returns S2l = (to_mult-1)/2"""
        self._i_need(("to_mult",))

        return float(self["to_mult"]-1)/2

    def get_deltak(self):
        """***Convention*** returns the Kronecker delta delta_K(0, Lambdal+Lambda2l)

        delta_K(0, Lambdal+Lambda2l) =

            1 if Lambda1+Lambda2 == 0 ([_ _ Sigma - _ _ Sigma ] transitions)

            0 otherwise

        This replaces using table pfantmol.cro field for the Honl-London factor normalization
        """
        return 1. if self["from_spdf"]+self["to_spdf"] == 0 else 0.

    def populate_parse_str(self, string):
        """
        Populates (from_*) and (to_*) taking PFANT molecule description string as input

        Args:
            string: str following convention below

        **Convention** formula-system string (FSS): "(formula) (ignored) [(system string)]".
                       Formula at start of string, followed by ignored characters until square
                       bracket.

        For the *system string* convention, see f311.filetypes.parse_system_str()

        FSS examples:

            "OH [A 2 Sigma - X 2 Pi]"

            "12C16O INFRARED [X 1 SIGMA+]"
        """

        fieldnames = ["from_label", "from_mult", "from_spdf", "to_label", "to_mult", "to_spdf"]

        # Formula
        symbols = basic.description_to_symbols(string)
        if symbols is not None:
            self["formula"] = basic.symbols_to_formula(symbols)
        else:
            match = re.search("[A-Za-z]+", string)
            if match is not None:
                self["formula"] = match.group()

        if self["formula"] is None:
            raise RuntimeError("Cannot determine formula from '{}'".format(string))

        # System
        self.update(zip(fieldnames, basic.parse_system_str(string)))

        # PFANT name and notes
        name, _, notes = basic.split_molecules_description(string)
        self["pfant_name"] = name
        self["pfant_notes"] = notes


    def populate_all_using_system_row(self, db, row):
        print(row["id_molecule"])

        formula = db.get_conn().execute("select formula from molecule where id = ?",
                                        (row["id_molecule"],)).fetchone()["formula"]

        self.populate_all_using_str(db,
            "{} [{}]".format(formula, pyfant.molconsts_to_system_str(row, pyfant.SS_PLAIN)))

    def populate_all_using_str(self, db, string):
        assert isinstance(db, pyfant.FileMolDB)

        self.populate_parse_str(string)

        self.populate_ids(db)

        for key in self:
            if key.startswith("id_"):
                if self[key] is None:
                    raise RuntimeError("Cannot determine '{}'".format(key))

        self.populate_all_using_ids(db)

    def populate_all_using_ids(self, db, id_molecule=None, id_system=None, id_pfantmol=None,
                               id_statel=None, id_state2l=None):
        """
        Populates completely, given all necessary table ids

        Args:
            db: FileMolDB object
            id_molecule: id in table "molecule"
            id_system: id in table "system"
            id_pfantmol: id in table "pfantmol"
            id_statel: id in table "state" for the initial state
            id_state2l: id in table "state" for the final state

        Arguments id_* have a fallback, which is self[argument].


        """
        assert isinstance(db, pyfant.FileMolDB)

        if id_molecule is None: id_molecule = self["id_molecule"]
        if id_system is None: id_system = self["id_system"]
        if id_pfantmol is None: id_pfantmol = self["id_pfantmol"]
        if id_statel is None: id_statel = self["id_statel"]
        if id_state2l is None: id_state2l = self["id_state2l"]

        if id_molecule is None: raise ValueError("Both argument and dictionary value 'id_molecule' are None")
        if id_system is None: raise ValueError("Both argument and dictionary value 'id_system' are None")
        if id_pfantmol is None: raise ValueError("Both argument and dictionary value 'id_pfantmol' are None")
        if id_statel is None: raise ValueError("Both argument and dictionary value 'id_statel' are None")
        if id_state2l is None: raise ValueError("Both argument and dictionary value 'id_state2l' are None")


        # key name, id, table name, prefix for key in dictionary
        _map = [("id_molecule", id_molecule, "molecule", ""),
                ("id_system", id_system, "system", ""),
                ("id_pfantmol", id_pfantmol, "pfantmol", ""),
                ("id_statel", id_statel, "state", "statel_"),
                ("id_state2l", id_state2l, "state", "state2l_"),
                ]

        for keyname, id_, tablename, prefix in _map:
            if id_ is not None:
                ti = db.get_table_info(tablename)
                row = db.get_conn().execute("select * from {} where id = ?".format(tablename),
                                            (id_,)).fetchone()

                if row is None:
                    raise ValueError("Invalid {}: id {} does not exist in table '{}'".format(keyname, id_, tablename))

                for fieldname, fieldspecs in ti.items():
                    if not fieldname.startswith("id"):
                        # 20230612: convert empty string ('') to None
                        value = row[fieldname]
                        if fieldspecs["type"] == "real" and isinstance(value, str) and value == "":
                            value = None


                        self[prefix + fieldname] = value

            self[keyname] = id_

    def populate_ids(self, db):
        """Populates (id_*) with values found using (molecule name) and (from_*) and (to_*)."""

        assert isinstance(db, pyfant.FileMolDB)

        ff = ("id_molecule", "id_system", "id_pfantmol", "id_statel", "id_state2l")
        methods = (self._populate_id_molecule, self._populate_id_system,
                   self._populate_id_pfantmol, self._populate_ids_state)

        for f in ff:
            self[f] = None

        for m in methods:
            try:
                m(db)
            except MolConstPopulateError:
                pass

    def _populate_id_molecule(self, db):
        self._i_need(("formula",))

        one = db.get_conn().execute("select id from molecule where formula = ?",
                                    (self["formula"],)).fetchone()
        self["id_molecule"] = one["id"] if one is not None else None

    def _populate_id_system(self, db):
        self._i_need(("id_molecule", "from_label", "from_mult", "from_spdf", "to_label", "to_mult", "to_spdf"))

        self["id_system"] = db.find_id_system(self)

    def _populate_ids_state(self, db):
        """Populates id_statel and id_state2l using (from, to)X(label, mult, spdf)"""

        self._i_need(("id_molecule", "from_label", "from_mult", "from_spdf", "to_label", "to_mult", "to_spdf"))

        _map = [("from", "id_statel"), ("to", "id_state2l")]
        for fromto, idfieldname in _map:
            fromto_ = (
            self["id_molecule"], self["{}_label".format(fromto)], self["{}_mult".format(fromto)],
            self["{}_spdf".format(fromto)])
            one = db.get_conn().execute(
                "select id from state where id_molecule = ? and label = ? and mult = ? and spdf = ?",
                fromto_).fetchone()

            self[idfieldname] = one["id"] if one is not None else None

    def _populate_id_pfantmol(self, db):
        """Populates id_pfantmol using formula, to_*, and from_*"""

        need = ("formula", "from_label", "from_mult", "from_spdf", "to_label", "to_mult", "to_spdf")
        if any((self[x] is None for x in need)):
            raise MolConstPopulateError("I need ({})".format(", ".join(need)))

        self["id_pfantmol"] = None

        # Will check match one by one. Will have to parse PFANT molecule descriptions.
        # **Note** currently not accounting for ambiguous match (always takes first match).
        #
        cursor = db.get_conn().execute("select * from pfantmol")
        for row in cursor:
            another = MolConsts()
            another.populate_parse_str(row["description"])

            found = True
            for field_name in need:
                if self[field_name] != another[field_name]:
                    found = False
                    break

            if found:
                self["id_pfantmol"] = row["id"]
                return



    def _i_need(self, need):
        if any((self[fieldname] is None for fieldname in need)):
            for fieldname in need:
                if self[fieldname] is None:
                    break

            method_name = inspect.stack()[1][3]
            raise MolConstPopulateError("Method {}() needs (), but at least '{}' is None".format(
                method_name, need, fieldname))


def some_molconsts(db=None, s="OH [A 2 SIGMA - X 2 PI]"):
    """
    Returns a MolConsts object completely populated

    Args:
        db: FileMolDB object.

            **Note** If None, this function will attempt to create 'moldb.xxxx.sqlite',
                     then attempt to delete it

        s: parseable formula-and-system (refer to
    """

    flag_newdb = False
    if db is None:
        flag_newdb = True
        db = pyfant.FileMolDB()
        db.init_default()

    ret = MolConsts()
    try:
        ret.populate_all_using_str(db, s)

    finally:
        db.get_conn().close()
        if flag_newdb:
            try:
                os.unlink(db.filename)
            except FileNotFoundError:
                pass

    ret.None_to_zero()

    return ret


class MolConstPopulateError(Exception):
    pass





#!/usr/bin/env python

"""
Compares two molecular lines files and prints report with differences
"""

import a99
import tabulate
import logging
import argparse
import pyfant


a99.logging_level = logging.INFO
a99.flag_log_file = True


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
     description=__doc__,
     formatter_class=a99.SmartFormatter
     )
    parser.add_argument('fn1', type=str, help='input file name', nargs=1)
    parser.add_argument('fn2', type=str, help='output file name', nargs=1)

    args = parser.parse_args()

    fn1 = args.fn1[0]
    fn2 = args.fn2[0]

    fa = pyfant.FileMolecules()
    fa.load(fn1)

    fb = pyfant.FileMolecules()
    fb.load(fn2)

    data = []
    header = [fn1, fn2]


    def append_data(objname, attrname, value0, value1):
        """Appends difference"""
        data.append(["{}.{} = {}".format(objname, attrname, value0),
                     "{}.{} = {}".format(objname, attrname, value1)])

    def check_append(objname, attrname, obj0, obj1):
        """Checks if attributes match and add data entry otherwise"""
        value0 = obj0.__getattribute__(attrname)
        value1 = obj1.__getattribute__(attrname)
        if value0 != value1:
            append_data(objname, attrname, value0, value1)

    check_append("file", "num_lines", fa, fb)

    to_check = ["formula", "nv", "num_lines", "fe", "do", "mm", "am", "bm", "ua", "ub", "te",
                "cro", "s", ]

    for i, (ma, mb) in enumerate(zip(fa, fb)):
        for attr in to_check:
            check_append("mol[{}]".format(i), attr, ma, mb)

    print(tabulate.tabulate(data, header))



#!/usr/bin/env python

"""PFANT command-line options file editor."""

import sys
import argparse
import logging
import a99
import pyfant
import os

a99.logging_level = logging.INFO
a99.flag_log_file = True

CLS_FILE = pyfant.FileOptions
CLS_FORM = pyfant.XFileOptions

if __name__ == "__main__":

    parser = argparse.ArgumentParser(description=__doc__, formatter_class=a99.SmartFormatter)
    parser.add_argument('fn', type=str, help="{} file name".format(a99.get_obj_doc0(CLS_FILE)),
                        default=CLS_FILE.default_filename, nargs='?')
    args = parser.parse_args()

    app = a99.get_QApplication([])
    form = CLS_FORM()

    if not os.path.isfile(args.fn):
        a99.get_python_logger().info("File not found: '{}'".format(args.fn))
    else:
        form.load_filename(args.fn)

    form.show()
    sys.exit(app.exec_())



from f311 import FilePyConfig
import a99


__all__ = ["FileConfigConvMol"]


@a99.froze_it
class FileConfigConvMol(FilePyConfig):
    """Configuration file for molecular lines conversion GUI (Python code)"""

    default_filename = "configconvmol.py"
    attrs = ["obj"]
    editors = []

    # Name of variable in module
    modulevarname = "ccm"



"""
Miscellanea routines that depend on other pyfant modules.

Rule: only 'gui/' modules can import util!!!

"""
import os
import glob
import shutil
import pyfant
import a99
import f311
import sqlite3
import unicodedata

__all__ = ["run_parallel", "setup_inputs", "copy_star", "link_to_data",
           "create_or_replace_or_skip_links", "copy_or_skip_files", "insert_states_from_nist",
           "insert_fcfs", "insert_molecule_from_nist", "run_traprb", ]

# ##################################################################################################
# Terminal-based interface


def run_parallel(rr, max_simultaneous=None, flag_console=False, runnable_manager=None,
                 flag_verbose=False, flag_exit_if_fail=False):
    """
    Args:
        rr: list of Runnable instances
        max_simultaneous: (defaults to RunnableManager.max_simultaneous)
            maximum number of simultaneous processes. **Note** ineffective if runnable_manager is passed
        flag_console: if True, will display menu with options, otherwise will exit when all runnables finish
        runnable_manager: if passed, will use passed; if not, will create new.
        flag_verbose: whether of not to log any messages (besides console) **Note** if runnable_manager is passed, it will keep its own
        flag_exit_if_fail: exit as soon as possible if any runnable fails? **Note** ineffective if runnable_manager is passed

    Returns: the RunnableManager object
    """
    import pyfant

    # Adds to pool
    logger = a99.get_python_logger()
    if runnable_manager:
        assert isinstance(runnable_manager, pyfant.RunnableManager)
        rm = runnable_manager
    else:
        rm = pyfant.RunnableManager(max_simultaneous=max_simultaneous, flag_verbose=flag_verbose,
                                flag_exit_if_fail=flag_exit_if_fail)
    flag_had_to_start = False
    if not rm.flag_start_called:
        rm.start()
        flag_had_to_start = True

    rm.add_runnables(rr)

    # Primitive thread monitor
    if flag_console:
        while True:
            print(("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%" + (
            " ALIVE" if rm.is_alive() else " DEAD")))
            print(rm)
            print(("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%" + (
            " ALIVE" if rm.is_alive() else " DEAD")))
            s = input("[Enter] -- [e]xit & keep in loop -- [q]uit -- [k]ill running >>> ")
            if s.lower() == "q":
                if rm.is_alive():
                    try:
                        rm.exit()
                    except:
                        logger.exception("Error trying to exit")
                break
            if s.lower() == "e":
                try:
                    rm.exit()
                except:
                    logger.exception("Error trying to exit")
            if s.lower() == "k":
                rm.kill_runnables()
    else:
        rm.wait_until_finished()
        if flag_had_to_start:
            rm.exit()

    if flag_verbose:
        a99.get_python_logger().info(
            ("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%" + (" ALIVE" if rm.is_alive() else " DEAD")))
        a99.get_python_logger().info("test-tm2 [SUPPOSED TO HAVE] EXITED")
        a99.get_python_logger().info(rm)

    return rm


def setup_inputs(dest_dir='.', star='sun-asplund-2009', common='common', h=True, atoms=True,
                 molecules=True, opa=True):
    """
    Sets up input data for spectral synthesis.

    Args:
      dest_dir='.': directory where files and links will be created
      star='sun-asplund-2009': directory (relative to PFANT/data) for stellar data
      common='common': directory (relative to PFANT/data) for star-independent data files
      h=True: whether to look for hmap.dat
      atoms=True: whether to look for atoms.dat
      molecules=True: whether to look for molecules.dat
      opa=True: whether to look for grid.moo
    """
    import pyfant as pf

    logger = a99.get_python_logger()
    dd = pyfant.get_pfant_path("data")

    # Functions that return full path, given a filename, to ...
    fd = lambda filename: os.path.join(dest_dir, filename)  # ... Destination directory
    fs = lambda filename: os.path.join(dd, star, filename)  # ... Stellar data directory
    fc = lambda filename: os.path.join(dd, common, filename)  # ... Common data directory

    # ## main.dat not present
    if not os.path.isfile(fd("main.dat")):
        zz_mnbp = ["main.dat", "abonds.dat",
                   "modeles.mod"]  # files that must not be present if main.dat is not present

        for z in zz_mnbp:
            if os.path.isfile(fd(z)):
                raise RuntimeError("Found file '%s' in local directory."
                                   "If 'main.dat' is not present, files %s must also not exist." %
                                   (z, zz_mnbp[1:]))

    # ## Stellar data...
    zz = ["main.dat", "abonds.dat"]
    copy_or_skip_files([fs(z) for z in zz], dest_dir=dest_dir)

    # ## Common data...
    zz = ["absoru2.dat", "partit.dat", "grid.mod"]
    if opa: zz.append("grid.moo")
    if h: zz.append("hmap.dat")
    if atoms: zz.append("atoms.dat")
    if molecules: zz.append("molecules.dat")
    create_or_replace_or_skip_links([fc(z) for z in zz], dest_dir=dest_dir)


def copy_star(src_dir=None, starname=None):
    """
    Copies files such as main.dat and abonds.dat from PFANT/data/some_directory into local directory

    Args:
        src_dir: absolute path to directory containing PFANT star data files FileMain and FileAbonds
                 (and optionally FileDissoc files, which will be also copied if present)
        starname: if passed, will ignore src_dir and make it from starname instead, considering
                  starname as a subdirectory of PFANT/data/
    """
    star_classes = [pyfant.FileMain, pyfant.FileDissoc, pyfant.FileAbonds]

    if starname is not None:
        src_dir = os.path.join(pyfant.get_pfant_data_path(), starname)

    if src_dir is None and starname is None:
        raise ValueError("Supply either src_dir or starname")

    if not os.path.isdir(src_dir):
        raise ValueError("'{}' is not a valid directory".format(src_dir))

    a99.get_python_logger().debug("Will look inside directory %s" % src_dir)

    # makes list of files to analyse
    types = ('*.dat', '*.mod')
    ff = []
    for type_ in types:
        ff.extend(glob.glob(os.path.join(src_dir, type_)))

    copy_or_skip_files(ff)


def link_to_data(src_dir=None):
    """Creates symbolic links to files '*.dat', '*.mod', '*.moo' within src_dir, skipping certain types

    Args:
        src_dir: directory containing files to be linked to. If omitted, it defaults to
                 PFANT/data/common
    """

    if src_dir is None:
        src_dir = pyfant.get_pfant_path('data', 'common')

    a99.get_python_logger().debug("Will look inside directory %s" % src_dir)

    # makes list of files to analyse
    types = ('*.dat', '*.mod', '*.moo')
    ff = []
    for type_ in types:
        ff.extend(glob.glob(os.path.join(src_dir, type_)))

    create_or_replace_or_skip_links(ff)


def create_or_replace_or_skip_links(ff, dest_dir="."):
    """Creates a series of links given a list of target filepaths.

    Args:
        ff: list of full path to files
        dest_dir: destination directory

    It skips files of types FileMain, FileAbonds, FileDissoc, FileToH
    """
    for f in ff:
        name = os.path.split(f)[1]
        ptd = os.path.join(dest_dir, name)  # path to destination

        flag_skip = False
        a99.get_python_logger().info(("Considering file '%s' ..." % name))
        if os.path.isfile(ptd) and not os.path.islink(ptd):
            _print_skipped("file exists in local directory")
            flag_skip = True
        else:
            obj = f311.load_with_classes(f, [pyfant.FileMain, pyfant.FileAbonds, pyfant.FileDissoc, pyfant.FileToH])
            if obj is not None:
                _print_skipped("detected type %s" % obj.__class__.__name__)
                flag_skip = True
            else:
                obj = f311.load_with_classes(f, [pyfant.FileModBin])
                if obj is not None:
                    if len(obj) == 1:
                        _print_skipped("%s of only one record" % obj.__class__.__name__)
                        flag_skip = True

        if not flag_skip:
            try:
                if os.path.islink(ptd):
                    os.remove(ptd)
                    s_action = "replaced existing"
                else:
                    s_action = "created"
                a99.create_symlink(f, ptd)
                a99.get_python_logger().info("   ... %s link" % s_action)
            except Exception as e:
                a99.print_error("Error creating link: %s" % a99.str_exc(e))
                raise


def copy_or_skip_files(ff, dest_dir="."):
    """Copies a series of files, skipping those which already exist.

    Args:
      ff: list of full paths to files to be copied
      dest_dir: destination directory
    """

    for f in ff:
        name = os.path.split(f)[1]

        flag_skip = False
        a99.get_python_logger().info("Considering file '%s' ..." % name)
        if os.path.isfile(name):
            _print_skipped("file exists in local directory")
            flag_skip = True
        else:
            obj = f311.load_with_classes(f, [pyfant.FileMain, pyfant.FileAbonds, pyfant.FileDissoc])
            if obj is not None:
                pass
            else:
                _print_skipped("neither main, abonds, nor dissoc file")
                flag_skip = True

        if not flag_skip:
            try:
                shutil.copy(f, dest_dir)
                a99.get_python_logger().info("   ... file copied")
            except Exception as e:
                a99.print_error("Error copying file: %s" % str(e))


def _print_skipped(reason):
    """Standardized printing for when a file was skipped."""
    a99.get_python_logger().info("   ... SKIPPED ({0!s}).".format(reason))


def insert_fcfs(moldb, id_system, fcfs, flag_replace=False):
    """
    Inserts Franck-Condon Factors for given system

    :param moldb: FileMolDB
    :param id_system: int, system ID in 'system' table
    :param fcfs: {(vl, v2l): fcf, ...}, as stored in a FileTRAPRBOutput object
    :param flag_replace: if FCFs for system identified by 'id_system' already exists,
        I may either raise (flag_replace==False) or replace these FCFs
    :return: None
    """
    conn = moldb.get_conn()
    cursor = conn.cursor()
    assert isinstance(conn, sqlite3.Connection)
    assert isinstance(cursor, sqlite3.Cursor)

    n = cursor.execute("select count(*) from fcf where id_system = ?",
                       (id_system,)).fetchone()["count(*)"]
    if n > 0:
        if flag_replace:
            cursor.execute("delete from fcf where id_system = ?", (id_system,))
        else:
            raise RuntimeError("FCFs exist for system #{}".format(id_system))

    for (vl, v2l), fcf in fcfs.items():
        conn.execute("insert into fcf (id_system, vl, v2l, value) values (?,?,?,?)",
                     (id_system, vl, v2l, fcf))

    conn.commit()


def insert_molecule_from_nist(moldb, formula, flag_replace=False, flag_do_what_i_can=True):
    """
    Retrieves NIST data and inserts new molecule into DB, or updates states if molecule exists

    :param moldb: FileMolDB object
    :param formula: str, e.g., "OH"
    :param flag_replace: applies to states (not molecule entry); passed on to insert_states_from_nist()
    :param flag_do_what_i_can: if set, will insert molecule even if cannot
        download molecular constants from NIST
    :return: None
    """
    conn = moldb.get_conn()
    cursor = conn.cursor()
    assert isinstance(conn, sqlite3.Connection)
    assert isinstance(cursor, sqlite3.Cursor)

    id_molecule = None
    data = None
    name = "?name?"
    n = conn.execute("select count(*) from molecule where formula = ?", (formula,)).fetchone()["count(*)"]
    if n > 0:
        id_molecule = conn.execute("select id from molecule where formula = ?", (formula,))
    else:
        try:
            data, header, name = pyfant.get_nist_webbook_constants(formula, flag_unicode=True,
                                                                   flag_parse_state=True)
        except:
            if flag_do_what_i_can:
                a99.get_python_logger().exception("Could not get NIST constants for formula '{}'".format(formula))
            else:
                raise

    if id_molecule is None:
        cursor.execute("insert into molecule (formula, name) values (?, ?)", (formula, name))
        id_molecule = cursor.lastrowid
        conn.commit()

    if data is not None:
        insert_states_from_nist(moldb, id_molecule, data, flag_replace)

    conn.commit()


def insert_states_from_nist(moldb, id_molecule, nist_data, flag_replace=False):
    """
    Inserts downloaded NIST data into table 'state'

    :param moldb: FileMolDB
    :param id_molecule: int, molecule ID in 'molecule' table
    :param nist_data: list of lists as returned by
                      get_nist_webbook_constants(..., flag_parse_state=True)
    :param flag_replace: if states for molecule identified by 'id_molecule' already exists,
        I may either raise (flag_replace==False) or replace these states
    :return: None
    """

    conn = moldb.get_conn()
    cursor = conn.cursor()
    assert isinstance(conn, sqlite3.Connection)
    assert isinstance(cursor, sqlite3.Cursor)

    n = cursor.execute("select count(*) from state where id_molecule = ?",
                       (id_molecule,)).fetchone()["count(*)"]
    if n > 0:
        if flag_replace:
            cursor.execute("delete from state where id_molecule = ?", (id_molecule,))
        else:
            formula = cursor.execute("select formula from molecule where id = ?",
                                     (id_molecule,)).fetchone()["formula"]
            raise RuntimeError("States exist for formula '{}'".format(formula))


    for state in nist_data:
        # **Note** assumes that the columns in data match the
        # (number of columns in the state table - 3) and their order
        conn.execute("insert into state values (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)",
                     [None, id_molecule] + state + [""])

    conn.commit()


def run_traprb(molconsts, fn_input=None, fn_output=None, maxv=None,
               ni=None, ns=None, igraph=None, ienerg=None, istate=None,
               rmin=None, rmax=None, delr=None):
    """
    Runs TRAPRB and returns a the TRAPRB object

    :param molconsts: MolConsts object
    :param fn_input: input filename (will be created with this name). If not passed, will make up a
                     filename as 'temp-in-*'
    :param fn_output: output filename (will be created with this name). If not passed, will make up
                     a filename as 'temp-out-*'
    :return: TRAPRB object

    For meanings of parameters and their default values, see filetraprb.py::TRAPRBInputState class

    **This routine creates files 'temp-traprb*' and leaves then**

    **Not multiprocessing-safe**, unless fn_input and fn_output are controlled externally.
    """

    if fn_input is None:
        fn_input = a99.new_filename("temp-in-")
    if fn_output is None:
        fn_output = a99.new_filename("temp-out-")

    fin = pyfant.FileTRAPRBInput()
    fin.from_molconsts(molconsts, maxv, ni, ns, igraph, ienerg, istate, rmin, rmax, delr, )
    fin.save_as(fn_input)

    r = pyfant.TRAPRB()
    r.fn_input = fn_input
    r.fn_output = fn_output
    r.run()

    return r




#!/usr/bin/env python

"""
Merges several PFANT molecular lines file into a single one
"""

import pyfant
import argparse
import os
import glob
import sys
import logging
import a99


a99.logging_level = logging.INFO
a99.flag_log_file = True


_AUTO = "(automatic)"
_PREFIX = "molecules-merged-"


def main(fns_input, fn_output):
    ffmol = []
    for fn in fns_input:
        fmol = pyfant.FileMolecules()
        try:
            fmol.load(fn)
            print("File '{}': {} molecule{}".format(
                fn,
                len(fmol),
                "" if len(fmol) == 1 else "s"
            ))
            ffmol.append(fmol)
        except:
            a99.get_python_logger().exception("Skipping file '{}'".format(fn))

    n = len(ffmol)

    fout = pyfant.FileMolecules()
    fout.titm = "Merge of {} file{}: {}".format(
        n,
        "s" if n != 1 else "",
        ", ".join([os.path.split(fmol.filename)[1] for fmol in ffmol]))

    for fmol in ffmol:
        fout.molecules.extend(fmol.molecules)
    print("Number of molecules in output: {}".format(len(fout)))

    if fn_output is None:
        fn_output = a99.new_filename(_PREFIX, ".dat")

    fout.save_as(fn_output)
    print("Saved file '{}'".format(fn_output))

if __name__ == "__main__":
    lggr = a99.get_python_logger()

    parser = argparse.ArgumentParser(description=__doc__, formatter_class=a99.SmartFormatter)
    parser.add_argument("files", type=str, nargs="+", help="files specification: list of files, wildcards allowed")
    parser.add_argument("-o", "--fn_output", nargs="?", type=str, default=_AUTO,
     help="output filename. If not specified, creates file such as '{}.0000.dat'".format(_PREFIX))

    args = parser.parse_args()

    # Compiles list of file names.
    # Each item in args.fn list may have wildcards, and these will be expanded
    # into actual filenames, then duplicates are eliminated
    patterns = args.files
    _ff = []
    for pattern in patterns:
        _ff.extend(glob.glob(pattern))
    ff = []
    for f in _ff:
        if f not in ff:
            ff.append(f)

    # Output
    fn_output = None if args.fn_output == _AUTO else args.fn_output

    main(ff, fn_output)
